from __future__ import print_function, division, absolute_import

import os
import sys
# Each process only run 1 threads
os.environ['MKL_NUM_THREADS'] = '1'
os.environ['NUMEXPR_NUM_THREADS'] = '1'
os.environ['OMP_NUM_THREADS'] = '1'

import time
import types
import pickle
import inspect
from six import add_metaclass
from decorator import decorator
from collections import defaultdict
from abc import ABCMeta, abstractmethod
from multiprocessing.pool import ThreadPool, Pool
from multiprocessing import (cpu_count, Process, Queue, Value, Lock,
                             current_process, Pipe)

import numpy as np

# ===========================================================================
# Threading
# ===========================================================================
_async_function_counter = defaultdict(int)
_NTHREADS = max(int(cpu_count()) // 2, 2)
_THREAD_POOL = None
_MAX_PIPE_BLOCK = 8 * 1024 * 1024 # 8MB
# this keep track all the running async-tasks that haven't
# finished
_RUNNING_ASYNC_TASKS = []

def set_max_threads(n):
  """ Set the maximum number of Threads for the
  ThreadPool that execute the async_task.
  """
  global _NTHREADS
  if _NTHREADS != n:
    _NTHREADS = n
    # re-init thread pool
    global _THREAD_POOL
    if _THREAD_POOL is not None:
      _THREAD_POOL.join()
      _THREAD_POOL.close()
    _THREAD_POOL = ThreadPool(processes=_NTHREADS)

def _full_func_name(f):
  return f.__module__ + "." + f.__class__.__qualname__ + '.' + f.__name__

class _async_task(object):
  """ A class converting blocking functions into
  asynchronous functions by using threads or processes.

  """

  def __init__(self, func, *args, **kw):
    self._is_threading = kw.pop('_is_threading')
    _async_function_counter[func] = _async_function_counter[func] + 1
    # specific name
    name = '<async_task:[%s]<%s|#%s>%s Finished:False>' % \
    ('Thread' if self._is_threading else 'Process',
        _full_func_name(func), _async_function_counter[func],
        '(%s)' % ';'.join(inspect.signature(func).parameters.keys()),
    )
    self._name = name
    self._func = func
    # check initialized pool
    # ====== using threading ====== #
    if self._is_threading:
      global _THREAD_POOL
      if _THREAD_POOL is None:
        _THREAD_POOL = ThreadPool(processes=_NTHREADS)
      self._async_task = _THREAD_POOL.apply_async(
          func=func, args=args, kwds=kw)
    # ====== using Multiprocessing ====== #
    else:
      def wrapped_func(conn, *args, **kwargs):
        results = self._func(*args, **kwargs)
        ret_data = pickle.dumps(results,
                                protocol=pickle.HIGHEST_PROTOCOL)
        ret_length = len(ret_data)
        conn.send(ret_length)
        sent_bytes = 0
        while sent_bytes < ret_length:
          data = ret_data[sent_bytes: sent_bytes + _MAX_PIPE_BLOCK]
          conn.send(data)
          sent_bytes += _MAX_PIPE_BLOCK

      parent_conn, child_conn = Pipe()
      self._async_task = Process(target=wrapped_func,
                                 args=(child_conn,) + tuple(args),
                                 kwargs=kw)
      self._async_task.start()
      self._conn = parent_conn

  @property
  def name(self):
    return self._name.replace(':False>', ':%s>' % bool(self.finished))

  def __str__(self):
    return self.name

  def __repr__(self):
    return str(self)

  @property
  def finished(self):
    if self._is_threading:
      return self._async_task.ready() and self._async_task.successful()
    else:
      return self._async_task.is_alive()

  @property
  def finish(self):
    return self.get()

  @property
  def result(self):
    return self.get()

  def get(self, timeout=None):
    """Return actual result of the function"""
    if not hasattr(self, '_result'):
      # ====== multi-threading ====== #
      if self._is_threading:
        self._result = self._async_task.get(timeout=timeout)
      # ====== multi-processing ====== #
      else:
        while not self._conn.poll(): # waiting for data
          pass
        ret_length = self._conn.recv()
        read_bytes = 0
        received = []
        while read_bytes < ret_length:
          data = self._conn.recv()
          received.append(data)
          read_bytes += len(data)
        received = b"".join(received)
        self._result = pickle.loads(received)
        # clean up
        self._conn.close()
        self._async_task.join(timeout=timeout)
    return self._result

def async(func=None):
  """ This create and asynchronized result using Threading instead of
  multiprocessing, you can take advantage from share memory in threading
  for async-IO using this decorator.

  Parameters
  ----------
  func: call-able
      main workload executed in this function and return the
      results.

  callback: call-able
      a callback function triggered when the task finished

  Example
  -------
  >>> from odin.utils import async
  ...
  >>> def test_fn(idx):
  ...   path = '/tmp/tmp%d.txt' % idx
  ...   with open(path, 'w') as f:
  ...     for i in range(100000):
  ...       f.write(str(i))
  ...     print("FINISH!", path)
  ...
  >>> f = async(test_fn)
  >>> x1 = f(1)
  >>> x2 = f(2)
  >>> x3 = f(3)
  ...
  >>> count = 0
  >>> while not x1.finished or not x2.finished or not x3.finished:
  ...   count += 1
  ...   print("Iteration:", count)
  ...   print(x1)
  ...   print(x2)
  ...   print(x3)

  """
  @decorator
  def _decorator_func_(func, *args, **kwargs):
    kwargs['_is_threading'] = True
    task = _async_task(func, *args, **kwargs)
    return task
  # roles are not specified, given function directly
  if inspect.isfunction(func) or inspect.ismethod(func):
    return _decorator_func_(func)
  # roles are specified
  return _decorator_func_

def async_mpi(func=None):
  """ This create and asynchronized result using multi-Processing
  Also check: `odin.utils.mpi.async` for multi-Threading
  decorator

  Parameters
  ----------
  func: call-able
      main workload executed in this function and return the
      results.

  callback: call-able
      a callback function triggered when the task finished

  Example
  -------
  >>> from odin.utils import async
  ...
  >>> def test_fn(idx):
  ...   path = '/tmp/tmp%d.txt' % idx
  ...   with open(path, 'w') as f:
  ...     for i in range(100000):
  ...       f.write(str(i))
  ...     print("FINISH!", path)
  ...
  >>> f = async(test_fn)
  >>> x1 = f(1)
  >>> x2 = f(2)
  >>> x3 = f(3)
  ...
  >>> count = 0
  >>> while not x1.finished or not x2.finished or not x3.finished:
  ...   count += 1
  ...   print("Iteration:", count)
  ...   print(x1)
  ...   print(x2)
  ...   print(x3)

  """
  @decorator
  def _decorator_func_(func, *args, **kwargs):
    kwargs['_is_threading'] = False
    task = _async_task(func, *args, **kwargs)
    return task
  # roles are not specified, given function directly
  if inspect.isfunction(func) or inspect.ismethod(func):
    return _decorator_func_(func)
  # roles are specified
  return _decorator_func_

# ===========================================================================
# Multi-processing
# ===========================================================================
_SLEEP_TIME = 0.01

def segment_list(l, size=None, n_seg=None):
  '''
  Example
  -------
  >>> segment_list([1,2,3,4,5],2)
  >>> # [[1, 2, 3], [4, 5]]
  >>> segment_list([1,2,3,4,5],4)
  >>> # [[1], [2], [3], [4, 5]]
  '''
  # by floor, make sure and process has it own job
  if n_seg is None:
    n_seg = int(np.ceil(len(l) / float(size)))
  # start segmenting
  segments = []
  start = 0
  remain_data = len(l)
  remain_seg = n_seg
  while remain_data > 0:
    # ====== adaptive size ====== #
    size = remain_data // remain_seg
    segments.append(l[start:(start + size)])
    # ====== update remain ====== #
    start += size
    remain_data -= size
    remain_seg -= 1
  return segments

class SharedCounter(object):
  """ A multiprocessing syncrhonized counter """

  def __init__(self, initial_value=0):
    self.val = Value('i', initial_value)
    self.lock = Lock()

  def add(self, value=1):
    with self.lock:
      self.val.value += value

  @property
  def value(self):
    with self.lock:
      return self.val.value

  def __del__(self):
    del self.lock
    del self.val

class MPI(object):
  """ MPI - Multi processing interface
  This class use round robin to schedule the tasks to each processes

  Parameters
  ----------
  jobs: list, tuple, numpy.ndarray
      list of works.
  func: call-able
      take a `list of jobs` as input (i.e. map_func([job1, job2, ...])),
      the length of this list is determined by `buffer_size`
      NOTE: the argument of map_func is always a list.
  ncpu: int
      number of processes.
  batch: int
      the amount of samples grouped into list, and feed to each
      process each iteration. (i.e. func([job0, job1, ...]))
      if `batch=1`, each input is feed individually to `func`
      (i.e. func(job0); func(job1]); ...)
  hwm: int
      "high water mark" for SEND socket, is a hard limit on the
      maximum number of outstanding messages ØMQ shall queue
      in memory for any single peer that the specified socket
      is communicating with.
  chunk_scheduler: bool
      if `True`, jobs are grouped into small chunks of `batch`, then each
      chunk will be feed to each process until the jobs are exhausted, hence,
      this approach guarantee that all processes will run until the end, but
      the execution speed will be lower since each processes need to
      continuously receives chunks from main process.
      if `False`, jobs are splited into equal size for each process at the
      beginning, do this if you sure all jobs require same processing time.
  backend: {'pyzmq', 'python'}
      using 'pyzmq' for interprocess communication or default python Queue.
  Note
  ----
  Using pyzmq backend often 3 time faster than python Queue
  """

  def __init__(self, jobs, func,
               ncpu=1, batch=1, hwm=144,
               backend='python'):
    super(MPI, self).__init__()
    backend = str(backend).lower()
    if backend not in ('pyzmq', 'python'):
      raise ValueError("Only support 2 backends: 'pyzmq', and 'python'")
    self._backend = backend
    self._ID = np.random.randint(0, 10e8, dtype=int)
    # ====== check map_func ====== #
    if not hasattr(func, '__call__'):
      raise Exception('"func" must be call-able')
    self._func = func
    # ====== MPI parameters ====== #
    # never use all available CPU
    if ncpu is None:
      ncpu = cpu_count() - 1
    self._ncpu = min(
        np.clip(int(ncpu), 1, cpu_count() - 1),
        len(jobs)
    )
    self._batch = max(1, int(batch))
    self._hwm = max(0, int(hwm))
    # ====== internal states ====== #
    self._nb_working_cpu = self._ncpu
    # processes manager
    self._is_init = False
    self._is_running = False
    self._is_finished = False
    self._terminate_now = False
    # ====== other queue ====== #
    if not isinstance(jobs, (tuple, list, np.ndarray)):
      raise ValueError("`jobs` must be instance of tuple or list.")
    self._jobs = jobs
    self._remain_jobs = SharedCounter(len(self._jobs))
    # Equally split for all processes
    self._tasks = Queue(maxsize=0)
    for i in segment_list(np.arange(len(self._jobs), dtype='int32'),
                          size=self._batch):
      self._tasks.put_nowait(i)
    for i in range(self._ncpu): # ending signal
      self._tasks.put_nowait(None)
    # ====== only 1 iteration is created ====== #
    self._current_iter = None

  # ==================== properties ==================== #
  def __len__(self):
    """ Return the number of remain jobs """
    return max(self._remain_jobs.value, 0)

  @property
  def nb_working_cpu(self):
    return self._nb_working_cpu

  @property
  def is_initialized(self):
    return self._is_init

  @property
  def is_finished(self):
    return self._is_finished

  @property
  def is_running(self):
    return self._is_running

  def terminate(self):
    self._terminate_now = True
    # force everything finished
    if self._current_iter is not None and \
    not self.is_finished:
      try:
        next(self._current_iter)
      except StopIteration:
        pass

  # ==================== helper ==================== #
  def __iter(self):
    # Initialize
    if not self._is_init:
      if self._backend == 'pyzmq':
        init_func = self._init_zmq
        run_func = self._run_pyzmq
      elif self._backend == 'python':
        init_func = self._init_python
        run_func = self._run_python
      init_func()
      self._is_init = True
    yield None # yeild not thing for init
    # Select run function
    self._is_running = True
    for i in run_func():
      if self._terminate_now:
        break
      yield i
    # Finalize
    self._is_running = False
    self._finalize()
    self._is_finished = True

  def __iter__(self):
    if self._current_iter is None:
      self._current_iter = self.__iter()
      next(self._current_iter)
    return self._current_iter

  # ==================== pyzmq ==================== #
  def _init_zmq(self):
    # this is ugly but work well
    import zmq

    # tasks_or_queue only return the indices, need to get it from self._jobs
    def wrapped_map(pID, tasks, remain_jobs):
      # ====== create ZMQ socket ====== #
      ctx = zmq.Context()
      sk = ctx.socket(zmq.PAIR)
      sk.set(zmq.SNDHWM, self._hwm)
      sk.set(zmq.LINGER, -1)
      sk.bind("ipc:///tmp/%d" % (self._ID + pID))

      # ====== Doing the jobs ====== #
      t = tasks.get()
      while t is not None:
        # `t` is just list of indices
        t = [self._jobs[i] for i in t]
        # monitor current number of remain jobs
        remain_jobs.add(-len(t))
        if self._batch == 1: # batch=1, NO need for list of inputs
          ret = self._func(t[0])
        else: # we have input is list of inputs here
          ret = self._func(t)
        # if a generator is return, traverse through the
        # iterator and return each result
        if not isinstance(ret, types.GeneratorType):
          ret = (ret,)
        for r in ret:
          # ignore None values
          if r is not None:
            sk.send_pyobj(r)
        # delete old data (this work, checked)
        del ret
        # ge tne tasks
        t = tasks.get()
      # ending signal
      sk.send_pyobj(None)
      # wait for ending message
      sk.recv()
      sk.close()
      ctx.term()
      sys.exit(0)
    # ====== start the processes ====== #
    self._processes = [Process(target=wrapped_map,
                               args=(i, self._tasks, self._remain_jobs))
                       for i in range(self._ncpu)]
    [p.start() for p in self._processes]
    # ====== pyzmq PULL socket ====== #
    ctx = zmq.Context()
    sockets = []
    for i in range(self._ncpu):
      sk = ctx.socket(zmq.PAIR)
      sk.set(zmq.RCVHWM, 0) # no limit receiving
      sk.connect("ipc:///tmp/%d" % (self._ID + i))
      sockets.append(sk)
    self._ctx = ctx
    self._sockets = sockets
    self._zmq_noblock = zmq.NOBLOCK
    self._zmq_again = zmq.error.Again

  def _run_pyzmq(self):
    while self._nb_working_cpu > 0:
      for sk in self._sockets:
        try:
          r = sk.recv_pyobj(flags=self._zmq_noblock)
          if r is None:
            self._nb_working_cpu -= 1
            sk.send(b'')
            sk.close()
            self._sockets.remove(sk)
          else:
            yield r
        except self._zmq_again:
          pass

  # ==================== python queue ==================== #
  def _init_python(self):
    def worker_func(tasks, queue, counter, remain_jobs):
      hwm = self._hwm
      minimum_update_size = max(hwm // self._ncpu, 1)
      # ====== Doing the jobs ====== #
      t = tasks.get()
      while t is not None:
        # `t` is just list of indices
        t = [self._jobs[i] for i in t]
        remain_jobs.add(-len(t)) # monitor current number of remain jobs
        if self._batch == 1: # batch=1, NO need for list of inputs
          ret = self._func(t[0])
        else: # we have input is list of inputs here
          ret = self._func(t)
        # if a generator is return, traverse through the
        # iterator and return each result
        if not isinstance(ret, types.GeneratorType):
          ret = (ret,)
        nb_returned = 0
        for r in ret:
          if r is not None: # ignore None values
            queue.put(r)
            nb_returned += 1
            # sometime 1 batch get too big, and we need to stop
            # putting too many data into the queue
            if nb_returned >= minimum_update_size:
              counter.add(nb_returned)
              nb_returned = 0
              while counter.value > hwm:
                time.sleep(_SLEEP_TIME)
        del ret # delete old data (this work, checked)
        # increase shared counter (this number must perfectly
        # counted, only 1 mismatch and deadlock will happen)
        if nb_returned > 0:
          counter.add(nb_returned)
        # check if we need to wait for the consumer here
        while counter.value > hwm:
          time.sleep(_SLEEP_TIME)
        # get new tasks
        t = tasks.get()
      # ending signal
      queue.put(None)
      sys.exit(0)
    # ====== multiprocessing variables ====== #
    self._queue = Queue(maxsize=0)
    self._counter = SharedCounter(initial_value=0)
    self._processes = [Process(target=worker_func,
                               args=(self._tasks, self._queue,
                                     self._counter, self._remain_jobs))
                       for i in range(self._ncpu)]
    [p.start() for p in self._processes]

  def _run_python(self):
    while self._nb_working_cpu > 0:
      r = self._queue.get()
      while r is None:
        self._nb_working_cpu -= 1
        if self._nb_working_cpu <= 0:
          break
        r = self._queue.get()
      if r is not None:
        self._counter.add(-1)
        yield r

  # ==================== finalize ==================== #
  def _finalize(self):
    # terminate or join all processes
    if self._terminate_now:
      [p.terminate() for p in self._processes
       if p._popen is not None and p.is_alive()]
    # only join started process which has _popen is not None
    else:
      [p.join() for p in self._processes
      if p._popen is not None]
    self._tasks.close()
    del self._remain_jobs
    # ====== pyzmq ====== #
    if self._backend == 'pyzmq':
      for sk in self._sockets:
        sk.close()
      self._ctx.term()
    # ====== python ====== #
    elif self._backend == 'python':
      self._queue.close()
      del self._counter



from __future__ import absolute_import, division, print_function

from enum import Flag, auto
from numbers import Number

import numpy as np
import tensorflow as tf
from tensorflow_probability import distributions as tfd

from odin.bay import distributions as obd


# ===========================================================================
# enum
# ===========================================================================
class Statistic(Flag):
  SAMPLE = auto()
  MEAN = auto()
  VAR = auto()
  STDDEV = auto()
  DIST = auto()


# ===========================================================================
# Logging
# ===========================================================================
def _dist2text(dist):
  return '[%s]%s batch:%s event:%s' % (dist.__class__.__name__, dist.dtype.name,
                                       dist.batch_shape, dist.event_shape)


def _extract_desc(dist, name, pad):
  text = pad + (name + ':' if len(name) > 0 else '') + _dist2text(dist) + '\n'
  obj_type = type(dist)
  for key in dir(dist):
    val = getattr(dist, key)
    if isinstance(val, tfd.Distribution) and \
      not isinstance(getattr(obj_type, key, None), property):
      text += _extract_desc(val, key, pad + ' ')
  return text


def print_dist(dist, return_text=False):
  assert isinstance(dist, tfd.Distribution)
  text = _extract_desc(dist, '', '')
  if return_text:
    return text[:-1]
  print(text)


# ===========================================================================
# Objectives
# ===========================================================================
def kl_divergence(q,
                  p,
                  use_analytic_kl=False,
                  q_sample=lambda q: q.sample(),
                  reduce_axis=(),
                  auto_remove_independent=True,
                  name=None):
  """ Calculating KL(q(x)||p(x))

  Parameters
  ----------
  q : tensorflow_probability.Distribution
    the posterior distribution

  p : tensorflow_probability.Distribution
    the prior distribution

  use_analytic_kl : bool (default: False)
    if True, use the close-form solution  for

  q_sample : {callable, Tensor, Number}
    callable for extracting sample from `q(x)` (takes q distribution
    as input argument)

  reudce_axis : {None, int, tuple}
    reduce axis when use MCMC to estimate KL divergence, default
    `()` mean keep all original dimensions

  auto_remove_independent : bool (default: True)
    if `q` or `p` is `tfd.Independent` wrapper, get the original
    distribution for calculating the analytic KL

  name : {None, str}

  Returns
  -------
  """
  if auto_remove_independent:
    # only remove Independent if one is Indepedent and another is not.
    if isinstance(q, tfd.Independent) and not isinstance(p, tfd.Independent):
      q = q.distribution
    if not isinstance(q, tfd.Independent) and isinstance(p, tfd.Independent):
      p = p.distribution
  q_name = [i for i in q.name.split('/') if len(i) > 0][-1]
  p_name = [i for i in p.name.split('/') if len(i) > 0][-1]
  with tf.compat.v1.name_scope(name, "KL_q%s_p%s" % (q_name, p_name)):
    if bool(use_analytic_kl):
      return tfd.kl_divergence(q, p)
    # using MCMC sampling for estimating the KL
    if callable(q_sample):
      z = q_sample(q)
    elif isinstance(q_sample, Number):
      z = q.sample(int(q_sample))
    else:
      z = q_sample
    # calculate the output, then perform reduction
    kl = q.log_prob(z) - p.log_prob(z)
    kl = tf.reduce_mean(input_tensor=kl, axis=reduce_axis)
    return kl



from __future__ import print_function, division, absolute_import

import numpy as np

# returns a list of augmented audio data, stereo or mono
def augment_audio(y, sr, n_augment = 0, allow_speedandpitch = True, allow_pitch = True,
    allow_speed = True, allow_dyn = True, allow_noise = True, allow_timeshift = True, tab="", quiet=False):

  mods = [y]                  # always returns the original as element zero
  length = y.shape[0]

  for i in range(n_augment):
    if not quiet:
      print(tab + "augment_audio: ", i + 1, "of", n_augment)
    y_mod = y
    count_changes = 0

    # change speed and pitch together
    if (allow_speedandpitch) and random_onoff():
      length_change = np.random.uniform(low=0.9, high=1.1)
      speed_fac = 1.0 / length_change
      if not quiet:
        print(tab + "    resample length_change = ", length_change)
      tmp = np.interp(np.arange(0, len(y), speed_fac), np.arange(0, len(y)), y)
      #tmp = resample(y,int(length*lengt_fac))    # signal.resample is too slow
      minlen = min(y.shape[0], tmp.shape[0])     # keep same length as original;
      y_mod *= 0                                    # pad with zeros
      y_mod[0:minlen] = tmp[0:minlen]
      count_changes += 1

    # change pitch (w/o speed)
    if (allow_pitch) and random_onoff():
      bins_per_octave = 24        # pitch increments are quarter-steps
      pitch_pm = 4                                # +/- this many quarter steps
      pitch_change = pitch_pm * 2 * (np.random.uniform() - 0.5)
      if not quiet:
        print(tab + "    pitch_change = ", pitch_change)
      y_mod = librosa.effects.pitch_shift(y, sr, n_steps=pitch_change, bins_per_octave=bins_per_octave)
      count_changes += 1

    # change speed (w/o pitch),
    if (allow_speed) and random_onoff():
      speed_change = np.random.uniform(low=0.9, high=1.1)
      if not quiet:
        print(tab + "    speed_change = ", speed_change)
      tmp = librosa.effects.time_stretch(y_mod, speed_change)
      minlen = min(y.shape[0], tmp.shape[0])        # keep same length as original;
      y_mod *= 0                                    # pad with zeros
      y_mod[0:minlen] = tmp[0:minlen]
      count_changes += 1

    # change dynamic range
    if (allow_dyn) and random_onoff():
      dyn_change = np.random.uniform(low=0.5, high=1.1)  # change amplitude
      if not quiet:
        print(tab + "    dyn_change = ", dyn_change)
      y_mod = y_mod * dyn_change
      count_changes += 1

    # add noise
    if (allow_noise) and random_onoff():
      noise_amp = 0.005 * np.random.uniform() * np.amax(y)
      if random_onoff():
        if not quiet:
          print(tab + "    gaussian noise_amp = ", noise_amp)
        y_mod += noise_amp * np.random.normal(size=length)
      else:
        if not quiet:
          print(tab + "    uniform noise_amp = ", noise_amp)
        y_mod += noise_amp * np.random.normal(size=length)
      count_changes += 1

    # shift in time forwards or backwards
    if (allow_timeshift) and random_onoff():
      timeshift_fac = 0.2 * 2 * (np.random.uniform() - 0.5)  # up to 20% of length
      if not quiet:
        print(tab + "    timeshift_fac = ", timeshift_fac)
      start = int(length * timeshift_fac)
      if (start > 0):
        y_mod = np.pad(y_mod, (start, 0), mode='constant')[0:y_mod.shape[0]]
      else:
        y_mod = np.pad(y_mod, (0, -start), mode='constant')[0:y_mod.shape[0]]
      count_changes += 1

    # last-ditch effort to make sure we made a change (recursive/sloppy, but...works)
    if (0 == count_changes):
      if not quiet:
        print("No changes made to signal, trying again")
      mods.append(augment_audio(y, sr, n_augment = 1, tab="      ", quiet=quiet)[1])
    else:
      mods.append(y_mod)

  return mods

""" scale frequency axis logarithmically """
def logscale_spec(spec, sr=44100, factor=20., alpha=1.0, f0=0.9, fmax=1):
  spec = spec[:, 0:256]
  timebins, freqbins = np.shape(spec)
  scale = np.linspace(0, 1, freqbins) #** factor

  # http://ieeexplore.ieee.org/xpl/login.jsp?tp=&arnumber=650310&url=http%3A%2F%2Fieeexplore.ieee.org%2Fiel4%2F89%2F14168%2F00650310
  scale = np.array(map(lambda x: x * alpha if x <= f0 else (fmax - alpha * f0) / (fmax - f0) * (x - f0) + alpha * f0, scale))
  scale *= (freqbins - 1) / max(scale)

  newspec = np.complex128(np.zeros([timebins, freqbins]))
  allfreqs = np.abs(np.fft.fftfreq(freqbins * 2, 1. / sr)[:freqbins + 1])
  freqs = [0.0 for i in range(freqbins)]
  totw = [0.0 for i in range(freqbins)]
  for i in range(0, freqbins):
    if (i < 1 or i + 1 >= freqbins):
      newspec[:, i] += spec[:, i]
      freqs[i] += allfreqs[i]
      totw[i] += 1.0
      continue
    else:
      # scale[15] = 17.2
      w_up = scale[i] - np.floor(scale[i])
      w_down = 1 - w_up
      j = int(np.floor(scale[i]))

      newspec[:, j] += w_down * spec[:, i]
      freqs[j] += w_down * allfreqs[i]
      totw[j] += w_down

      newspec[:, j + 1] += w_up * spec[:, i]
      freqs[j + 1] += w_up * allfreqs[i]
      totw[j + 1] += w_up

  for i in range(len(freqs)):
    if (totw[i] > 1e-6):
      freqs[i] /= totw[i]

  return newspec, freqs



from __future__ import print_function, division, absolute_import

import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = "cpu=1,float32"
import shutil

import numpy as np

from odin import fuel as F, nnet as N
from odin import preprocessing as pp
from odin.utils import (get_all_files, get_all_ext, exec_commands,
                        MPI, cpu_count, Progbar, ArgController,
                        stdio, ctext, crypto)

args = ArgController(
).add('path', "path to TIDIGITS dataset"
).add('--wav', "re-run Converting sphere file to wave", False
).add('--ds', "re-run Group wave files into a dataset", False
).add('--compress', "re-run compression of the dataset", False
).parse()

TOTAL_FILES = 25096
README = \
"""
Original sample rate: 20,000 Hz
Downsampled sample rate: 8,000 Hz

Saved WAV file format:
    * [train|test]
    * [m|w|b|g] (alias for man, women, boy, girl)
    * [age]
    * [dialectID]
    * [speakerID]
    * [production]
    * [digit_sequence]
    => "train_g_08_17_as_a_4291815"

    train material, child "g"irl, age is "08", dialect group is "17",
    speaker code "as", "a" is first production,
    digit sequence "4-2-9-1-8-1-5".
--------------------
Category       Symbol    Number    Age Range (years)
  Man            M        111           21 - 70
  Woman          W        114           17 - 59
  Boy            B         50            6 - 14
  Girl           G         51            8 - 15

Eleven digits were used:  "zero", "one", "two", ... , "nine", and "oh".

Digits for each speaker:
    22 isolated digits (two tokens of each of the eleven digits)
    11 two-digit sequences
    11 three-digit sequences
    11 four-digit sequences
    11 five-digit sequences
    11 seven-digit sequences

Example of original data:
     => "/data/adults/train/man/fd/6z97za.wav"
     training material, adult male, speaker code "fd",
     digit sequence "six zero nine seven zero", "a" is first production.

     => "/data/adults/test/woman/pf/1b.wav"
     test material, adult female, speaker code "pf",
     digit sequence "one", "b" is second production.
------------------
    City                      Dialect             M    W    B    G

01 Boston, MA            Eastern New England      5    5    0    1
02 Richmond, VA          Virginia Piedmont        5    5    2    4
03 Lubbock, TX           Southwest                5    5    0    1
04 Los Angeles, CA       Southern California      5    5    0    1
05 Knoxville, TN         South Midland            5    5    0    0
06 Rochester, NY         Central New York         6    6    0    0
07 Denver, CO            Rocky Mountains          5    5    0    0
08 Milwaukee, WS         North Central            5    5    2    0
09 Philadelphia, PA      Delaware Valley          5    6    0    1
10 Kansas City, KS       Midland                  5    5    4    1
11 Chicago, IL           North Central            5    5    1    2
12 Charleston, SC        South Carolina           5    5    1    0
13 New Orleans, LA       Gulf South               5    5    2    0
14 Dayton, OH            South Midland            5    5    0    0
15 Atlanta, GA           Gulf South               5    5    0    1
16 Miami, FL             Spanish American         5    5    1    0
17 Dallas, TX            Southwest                5    5   34   36
18 New York, NY          New York City            5    5    2    2
19 Little Rock, AR       South Midland            5    6    0    0
20 Portland, OR          Pacific Northwest        5    5    0    0
21 Pittsburgh, PA        Upper Ohio Valley        5    5    0    0
22                       Black                    5    6    1    1

                         Total Speakers         111  114   50   51    326
"""

# ===========================================================================
# CONST
# ===========================================================================
# ====== main path ====== #
# inpath = "/mnt/sdb1/TIDIGITS"
inpath = args.path
outpath = '/home/trung/data/TIDIGITS_wav'
compress_path = '/home/trung/data/TIDIGITS.zip'
# ====== others ====== #
wav_path = os.path.join(inpath, "wave")
infopath = os.path.join(inpath, 'data/children/doc/spkrinfo.txt')
logpath = os.path.join(inpath, 'log.txt')
print("Input path:       ", ctext(inpath, 'cyan'))
print("Output path:      ", ctext(outpath, 'cyan'))
print("Convert to WAV at:", ctext(wav_path, 'cyan'))
print("Log path:         ", ctext(logpath, 'cyan'))
stdio(logpath)

exts = get_all_ext(inpath)
audio_files = get_all_files(inpath,
                filter_func=lambda f: f[-4:] == '.wav' and
                            f.split('/')[-3] in ('girl', 'boy', 'man', 'woman'))
# ID     Gender     Age     Dialect    Usage
# ID - Unique 2-character speaker identifier
# Gender - (M-man, W-woman, B-boy, G-girl)
# Age - Speaker age at time of recording
# Dialect - Dialect region identifier (see file "dialects.txt" for decode)
# Usage - (TST-test material, TRN-training material)
info = np.genfromtxt(infopath, dtype=str, skip_header=12)
info = {ID.lower(): (Gender.lower(), Age, Dialect, Usage)
        for ID, Gender, Age, Dialect, Usage in info}
gender_map = {
    "man": "m",
    "woman": "w",
    "boy": "b",
    "girl": "g"
}
usage_map = {
    "TST": "test",
    "TRN": "train"
}

def get_name(path):
  # integrate all info into the name
  usage, gender, ID, digits = path.split('/')[-4:]
  production = digits[-5]
  digits = digits[:-5]
  gender = gender_map[gender]
  gender_, age_, dialect_, usage_ = info[ID]
  usage_ = usage_map[usage_]
  assert usage == usage_ and gender == gender_, path
  name = '%s_%s_%s_%s_%s_%s_%s.wav' % \
      (usage, gender, age_, dialect_, ID, production, digits)
  return name
# ===========================================================================
# Convert all SPHERE to wav using sph2pipe
# ===========================================================================
if os.path.exists(wav_path):
  if args.wav:
    print("Override wave files at '%s'" % wav_path)
    shutil.rmtree(wav_path)
  elif len(os.listdir(wav_path)) != TOTAL_FILES:
    print("Found only %d files at '%s', delete old wave files" %
      (len(os.listdir(wav_path)), wav_path))
    shutil.rmtree(wav_path)
# ====== convert all compress audio to .wav using sph2pipe ====== #
if not os.path.exists(wav_path):
  os.mkdir(wav_path)
  cmds = ["sph2pipe %s %s -f rif" % (path, os.path.join(wav_path, get_name(path)))
          for path in audio_files]

  def mpi_fn(cmd):
    exec_commands(cmd, print_progress=False)
    yield len(cmd)
  prog = Progbar(target=len(cmds),
                 print_report=True, print_summary=True,
                 name='Converting .sph to .wav')
  # run the MPI tasks
  mpi = MPI(jobs=cmds, func=mpi_fn,
            ncpu=cpu_count() - 1, batch=12)
  for i in mpi:
    prog.add(i)
# ===========================================================================
# Extract Acoustic features
# ===========================================================================
jobs = get_all_files(wav_path,
                     filter_func=lambda x: '.wav' == x[-4:])
assert len(jobs) == TOTAL_FILES
# ====== configuration ====== #
if not os.path.exists(outpath) or args.ds:
  extractors = pp.make_pipeline(steps=[
      pp.speech.AudioReader(sr=None, sr_new=8000, best_resample=True,
                            remove_dc=True),
      pp.base.Converter(converter=lambda x: os.path.basename(x).split('.')[0],
                        input_name='path', output_name='name'),
      pp.base.AsType(dtype='float16', input_name='raw')
  ], debug=False)
  processor = pp.FeatureProcessor(jobs=jobs, path=outpath, extractor=extractors,
                                  n_cache=0.08, ncpu=None, override=True)
  processor.run()
  pp.validate_features(processor, path='/tmp/tidigits', nb_samples=12,
                       override=True)
  with open(os.path.join(outpath, 'README'), 'w') as f:
    f.write(README)
# ====== check the preprocessed dataset ====== #
ds = F.Dataset(outpath, read_only=True)
print(ds)
print(ctext(ds.md5, 'yellow'))
ds.close()
# ====== compress ====== #
if not os.path.exists(compress_path) or args.compress:
  if os.path.exists(compress_path):
    os.remove(compress_path)
  crypto.zip_aes(in_path=outpath, out_path=compress_path,
                 verbose=True)
print("Log at path:", ctext(logpath, 'cyan'))



from __future__ import print_function, division, absolute_import
import os
os.environ['ODIN'] = 'float32,gpu'

import numpy as np

from odin import ml
from odin import fuel as F
from odin.utils import args_parse, ctext, stdio, Progbar

from utils import (get_model_path, prepare_ivec_data, csv2mat,
                   TRAIN_DATA)
# ===========================================================================
# Configs
# ===========================================================================
args = args_parse([
    ('recipe', 'the name of function defined in feature_recipes.py', None),
    ('-nmix', "Number of GMM mixture", None, 2048),
    ('-tdim', "Dimension of t-matrix", None, 600),
    ('-feat', "Acoustic feature", ('mspec', 'bnf'), 'bnf'),
    ('--gmm', "Force re-run training GMM", None, False),
    ('--stat', "Force re-extraction of centered statistics", None, False),
    ('--tmat', "Force re-run training Tmatrix", None, False),
    ('--ivec', "Force re-run extraction of i-vector", None, False),
    ('--all', "Run all the system again, just a shortcut", None, False),
])
args.gmm |= args.all
args.stat |= args.all | args.gmm
args.tmat |= args.all | args.stat
args.ivec |= args.all | args.tmat
FEAT = args.feat
EXP_DIR, MODEL_PATH, LOG_PATH, TRAIN_PATH, TEST_PATH = get_model_path('ivec', args)
stdio(LOG_PATH)
# ===========================================================================
# Load dataset
# ===========================================================================
X, train, test = prepare_ivec_data(args.recipe, FEAT)
# ===========================================================================
# Training I-vector model
# ===========================================================================
ivec = ml.Ivector(path=MODEL_PATH, nmix=args.nmix, tv_dim=args.tdim,
                  niter_gmm=16, niter_tmat=16,
                  downsample=2, stochastic_downsample=True,
                  device='gpu', name="VoxCelebIvec")
ivec.fit(X, indices=train,
         extract_ivecs=True, keep_stats=False)
# ====== extract train i-vector ====== #
I_train = F.MmapData(ivec.ivec_path, read_only=True)
name_train = np.genfromtxt(ivec.name_path, dtype=str)
print("Train i-vectors:", ctext(I_train, 'cyan'))
# save train i-vectors to csv
prog = Progbar(target=len(name_train),
               print_report=True, print_summary=True,
               name="Saving train i-vectors")
with open(TRAIN_PATH, 'w') as f_train:
  for i, name in enumerate(name_train):
    spk = TRAIN_DATA[name]
    vec = I_train[i]
    f_train.write('\t'.join([str(spk)] + [str(v) for v in vec]) + '\n')
    prog.add(1)
# ====== extract test i-vector ====== #
test = sorted(test.items(), key=lambda x: x[0])
I_test = ivec.transform(X, indices=test,
                        save_ivecs=False, keep_stats=False)
# save test i-vector to csv
with open(TEST_PATH, 'w') as f_test:
  for (name, (start, end)), z in zip(test, I_test):
    f_test.write('\t'.join([name] + [str(i) for i in z]) + '\n')
# ====== print the model ====== #
csv2mat(exp_dir=EXP_DIR)
print(ivec)



from __future__ import absolute_import, division, print_function

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import confusion_matrix as tf_cm

from odin.backend.tensor import dimshuffle, to_llr, to_nonzeros
from odin.utils import as_tuple, is_number


# ===========================================================================
# Losses
# ===========================================================================
def binary_accuracy(y_true,
                    y_pred,
                    threshold=0.5,
                    reduction=tf.reduce_mean,
                    name=None):
  """ Non-differentiable """
  with tf.name_scope(name, "binary_accuracy", [y_pred, y_true, threshold]):
    if y_pred.shape.ndims > 1:
      y_pred = tf.reshape(y_pred, (-1,))
    if y_true.shape.ndims > 1:
      y_true = tf.reshape(y_true, (-1,))
    y_pred = tf.greater_equal(y_pred, threshold)
    match_values = tf.cast(tf.equal(tf.cast(y_pred, 'int32'),
                                    tf.cast(y_true, 'int32')),
                           dtype='int32')
    return reduction(match_values)


def categorical_accuracy(y_true,
                         y_pred,
                         top_k=1,
                         reduction=tf.reduce_mean,
                         name=None):
  """ Non-differentiable """
  with tf.name_scope(name, "categorical_accuracy", [y_true, y_pred]):
    if y_true.shape.ndims == y_pred.shape.ndims:
      y_true = tf.argmax(y_true, axis=-1)
    elif y_true.shape.ndims != y_pred.shape.ndims - 1:
      raise TypeError('rank mismatch between y_true and y_pred')
    if top_k == 1:
      # standard categorical accuracy
      top = tf.argmax(y_pred, axis=-1)
      y_true = tf.cast(y_true, top.dtype.base_dtype)
      match_values = tf.equal(top, y_true)
    else:
      match_values = tf.nn.in_top_k(y_pred, tf.cast(y_true, 'int32'), k=top_k)
    match_values = tf.cast(match_values, dtype='float32')
    return reduction(match_values)


def confusion_matrix(y_true, y_pred, labels=None, normalize=False, name=None):
  """
  Computes the confusion matrix of given vectors containing
  actual observations and predicted observations.

  Parameters
  ----------
  y_true : 1-d or 2-d tensor variable
      true values
  y_pred : 1-d or 2-d tensor variable
      prediction values
  normalize : bool
      if True, normalize each row to [0., 1.]
  labels : array, shape = [nb_classes], int (nb_classes)
      List of labels to index the matrix. This may be used to reorder
      or select a subset of labels.
      If none is given, those that appear at least once
      in ``y_true`` or ``y_pred`` are used in sorted order.

  Note
  ----
  if you want to calculate: Precision, Recall, F1 scores from the
  confusion matrix, set `normalize=False`

  """
  with tf.name_scope(name, 'confusion_matrix', [y_true, y_pred]):
    nb_classes = None
    if y_true.shape.ndims == 2:
      nb_classes = y_true.shape.as_list()[-1]
      y_true = tf.argmax(y_true, -1)
    elif y_true.shape.ndims != 1:
      raise ValueError('actual must be 1-d or 2-d tensor variable')
    if y_pred.shape.ndims == 2:
      nb_classes = y_pred.shape.as_list()[-1]
      y_pred = tf.argmax(y_pred, -1)
    elif y_pred.shape.ndims != 1:
      raise ValueError('pred must be 1-d or 2-d tensor variable')
    # check valid labels
    if labels is None:
      if nb_classes is None:
        raise RuntimeError(
            "Cannot infer the number of classes for confusion matrix")
      labels = int(nb_classes)
    elif is_number(labels):
      labels = int(labels)
    elif hasattr(labels, '__len__'):
      labels = len(labels)
    # transpose to match the format of sklearn
    cm = tf_cm(labels=y_true, predictions=y_pred, num_classes=labels)
    if normalize:
      cm = tf.cast(cm, dtype='float32')
      cm = cm / tf.reduce_sum(cm, axis=1, keep_dims=True)
    return cm


def detection_matrix(y_true, y_pred):
  # TODO
  pass


# ===========================================================================
# Speech task metrics
# ===========================================================================
def compute_Cavg(y_llr,
                 y_true,
                 cluster_idx=None,
                 Ptrue=0.5,
                 Cfa=1.,
                 Cmiss=1.,
                 probability_input=False):
  ''' Fast calculation of Cavg (for only 1 clusters)

  Parameters
  ----------
  y_llr: (nb_samples, nb_classes)
      log likelihood ratio: llr = log (P(data|target) / P(data|non-target))
  y_true: numpy array of shape (nb_samples,)
      Class labels.
  cluster_idx: list,
      Each element is a list that represents a particular language
      cluster and contains all class labels that belong to the cluster.
  Ptar: float, optional
      Probability of a target trial.
  Cfa: float, optional
      Cost for False Acceptance error.
  Cmiss: float, optional
      Cost for False Rejection error.
  probability_input: boolean
      if True, `y_llr` is the output probability from softmax and perform
      llr transform for `y_llr`

  Returns
  -------
  cluster_cost: numpy array of shape (n_clusters,)
      It contains average percentage costs for each cluster as defined by
      NIST LRE-15 language detection task. See
      http://www.nist.gov/itl/iad/mig/upload/LRE15_EvalPlan_v22-3.pdf
  total_cost: float
      An average percentage cost over all clusters.

  '''
  if probability_input:
    y_llr = to_llr(y_llr)
  thresh = np.log(Cfa / Cmiss) - np.log(Ptrue / (1 - Ptrue))
  nb_classes = y_llr.shape[1].value
  if isinstance(y_true, (list, tuple)):
    y_true = np.asarray(y_true)
  if y_true.shape.ndims == 1:
    y_true = tf.one_hot(y_true, depth=nb_classes, axis=-1)
  y_true = tf.cast(y_true, y_llr.dtype.base_dtype)
  # ====== statistics ====== #
  # invert of y_true, False Negative mask
  y_false = 1. - y_true
  y_positive = tf.cast(tf.greater_equal(y_llr, thresh), y_llr.dtype.base_dtype)
  # invert of y_positive
  y_negative = tf.cast(tf.less(y_llr, thresh), y_llr.dtype.base_dtype)
  distribution = tf.clip_by_value(tf.reduce_sum(y_true, axis=0), 10e-8,
                                  10e8)  # no zero values
  # ====== Pmiss ====== #
  miss = tf.reduce_sum(y_true * y_negative, axis=0)
  Pmiss = 100 * (Cmiss * Ptrue * miss) / distribution
  # ====== Pfa ====== # This calculation give different results
  fa = tf.reduce_sum(y_false * y_positive, axis=0)
  Pfa = 100 * (Cfa * (1 - Ptrue) * fa) / distribution
  Cavg = tf.reduce_mean(Pmiss) + tf.reduce_mean(Pfa) / (nb_classes - 1)
  return Cavg


def compute_Cnorm(y_true,
                  y_score,
                  Ptrue=[0.1, 0.5],
                  Cfa=1.,
                  Cmiss=1.,
                  probability_input=False):
  """ Computes normalized detection cost function (DCF) given
    the costs for false accepts and false rejects as well as a priori
    probability for target speakers.

    * This is the actual cost, different from the min cost (minDCF)

    (By convention, the more positive the score,
    the more likely is the target hypothesis.)

  Parameter
  ---------
  y_true: {array [n_samples], or list of array}
      each array is labels of binary or multi-classes
      detection tasks, each array can be an array of
      classes indices, or one-hot-encoded matrix.
      If multiple array are given, calculating `equalized cost`
      of all partitions, an example of 2 partitions are:
      VAST and MLSR14 files
  y_score: {array [n_samples, n_classes], or list of array}
      the outputs scores, can be probabilities values or log-likelihood
      values by default, the
  Ptrue: float [0.,1.], or list of float
      hypothesized prior probabilities of positive class,
      you can given multiple values by providing an array
  Cfa: float
      weight for False Alarm - False Positive error
  Cmiss: float
      weight for Miss - False Negative error

  Return
  ------
  C_norm: array [len(Ptrue)]
      minimum detection cost accordingly for each given value of `Ptrue`.
  C_norm_array: array [len(Ptrue), n_classes]
      minimum detection cost for each class, accordingly to each
      given value of `Ptrue`

  """
  y_true = as_tuple(y_true, t=np.ndarray)
  y_score = as_tuple(y_score, t=np.ndarray)
  if len(y_true) != len(y_score):
    raise ValueError("There are %d partitions for `y_true`, but %d "
                     "partitions for `y_score`." % (len(y_true), len(y_score)))
  if len(set(i.shape[1] for i in y_score)) != 1:
    raise ValueError(
        "The number of classes among scores array is inconsistent.")
  nb_partitions = len(y_true)
  # ====== preprocessing ====== #
  y_true = [np.argmax(i, axis=-1) if i.ndim >= 2 else i for i in y_true]
  nb_classes = y_score[0].shape[1]
  # threshold
  Ptrue = np.asarray(as_tuple(Ptrue), dtype=float)
  nb_threshold = len(Ptrue)
  # log(beta) is threshold, i.e.
  # if Ptrue=0.5 => beta=1. => threshold=0.
  beta = (Cfa / Cmiss) * ((1 - Ptrue) / Ptrue)
  beta = np.clip(beta, a_min=np.finfo(float).eps, a_max=np.inf)
  # ====== Cavg ====== #
  global_cm_array = np.zeros(shape=(nb_threshold, nb_classes, nb_classes))
  # Apply threshold on the scores and compute the confusion matrix
  for scores, labels in zip(y_score, y_true):
    actual_TP_per_class = np.lib.arraysetops.unique(ar=labels,
                                                    return_counts=True)[1]
    if probability_input:  # special case input is probability values
      scores = to_llr(scores)
    for theta_ix, theta in enumerate(np.log(beta)):
      thresholded_scores = (scores > theta).astype(int)
      # compute confusion matrix, this is different from
      # general implementation of confusion matrix above
      cm = np.zeros(shape=(nb_classes, nb_classes), dtype=np.int64)
      for i, (trial, target) in enumerate(zip(thresholded_scores, labels)):
        cm[target, :] += trial
      # miss and fa
      predic_TP_per_class = cm.diagonal()
      # Compute the number of miss per class
      nb_miss_per_class = actual_TP_per_class - predic_TP_per_class
      cm_miss_fa = cm
      cm_miss_fa[np.diag_indices_from(cm)] = nb_miss_per_class
      cm_probabilities = cm_miss_fa / actual_TP_per_class[:, None]
      # update global
      global_cm_array[theta_ix] += cm_probabilities
  # normalize by partitions
  global_cm_array /= nb_partitions
  # Extract probabilities of false negatives from confusion matrix
  p_miss_arr = global_cm_array.diagonal(0, 1, 2)
  p_miss = p_miss_arr.mean(1)
  # Extract probabilities of false positives from confusion matrix
  p_false_alarm_arr = (global_cm_array.sum(1) - p_miss_arr) / (nb_classes - 1)
  p_false_alarm = p_false_alarm_arr.mean(1)
  # Compute costs per languages
  C_Norm_arr = p_miss_arr + beta[:, None] * p_false_alarm_arr
  # Compute overall cost
  C_Norm = p_miss + beta * p_false_alarm
  return C_Norm, C_Norm_arr


def compute_minDCF(Pfa, Pmiss, Cmiss=1, Cfa=1, Ptrue=0.5):
  """ Estimating the min value of the  detection
  cost function (DCF)

  Parameters
  ----------
  Pfa: array, [n_samples]
      false alarm rate or false positive rate
  Pmiss: array, [n_samples]
      miss rate or false negative rate
  Cmiss: scalar
      weight for false positive mistakes
  Cfa: scalar
      weight for false negative mistakes
  Ptrue: scalar [0., 1.]
      prior probability of positive cases.

  Return
  ------
  min_DCF: scalar
      minimum value of the detection cost function for
      a given detection error trade-off curve
  Pfa_optimum: scalar
      and false alarm trade-off probabilities.
  Pmiss_optimum: scalar
      the correcponding miss

  """
  assert Pmiss.shape == Pfa.shape
  Pfalse = 1 - Ptrue
  # detection cost function vector
  DCF_vector = (Cmiss * Pmiss * Ptrue) + \
      (Cfa * Pfa * Pfalse)
  # get the optimal value and corresponding index
  min_idx = np.argmin(DCF_vector)
  min_val = DCF_vector[min_idx]
  return min_val, Pfa[min_idx], Pmiss[min_idx]


def compute_EER(Pfa, Pmiss):
  """ computes the equal error rate (EER) given
  Pmiss or False Negative Rate
  and
  Pfa or False Positive Rate
  calculated for a range of operating points on the DET curve

  @Author: "Timothee Kheyrkhah, Omid Sadjadi"
  """
  fpr, fnr = Pfa, Pmiss
  diff_pm_fa = fnr - fpr
  x1 = np.flatnonzero(diff_pm_fa >= 0)[0]
  x2 = np.flatnonzero(diff_pm_fa < 0)[-1]
  a = (fnr[x1] - fpr[x1]) / (fpr[x2] - fpr[x1] - (fnr[x2] - fnr[x1]))
  return fnr[x1] + a * (fnr[x2] - fnr[x1])


def compute_AUC(x, y, reorder=False):
  """Compute Area Under the Curve (AUC) using the trapezoidal rule

  This is a general function, given points on a curve.  For computing the
  area under the ROC-curve, see :func:`roc_auc_score`.  For an alternative
  way to summarize a precision-recall curve, see
  :func:`average_precision_score`.

  Parameters
  ----------
  x : array, shape = [n]
      x coordinates.
  y : array, shape = [n]
      y coordinates.
  reorder : boolean, optional (default=False)
      If True, assume that the curve is ascending in the case of ties, as for
      an ROC curve. If the curve is non-ascending, the result will be wrong.

  Returns
  -------
  auc : float

  Examples
  --------
  >>> import numpy as np
  >>> from sklearn import metrics
  >>> y = np.array([1, 1, 2, 2])
  >>> pred = np.array([0.1, 0.4, 0.35, 0.8])
  >>> fpr, tpr, thresholds = metrics.roc_curve(y, pred, pos_label=2)
  >>> metrics.auc(fpr, tpr)
  0.75

  """
  from sklearn.metrics import auc
  return auc(x, y, reorder)


def roc_curve(y_true,
              y_score,
              pos_label=None,
              sample_weight=None,
              drop_intermediate=True):
  """Compute Receiver operating characteristic (ROC)

  @copy from sklearn for convenience

  Note: this implementation is restricted to the binary classification task.

  Parameters
  ----------

  y_true : array, shape = [n_samples]
      True binary labels in range {0, 1} or {-1, 1}.  If labels are not
      binary, pos_label should be explicitly given.
  y_score : array, shape = [n_samples]
      Target scores, can either be probability estimates of the positive
      class, confidence values, or non-thresholded measure of decisions
      (as returned by "decision_function" on some classifiers).
  pos_label : int or str, default=None
      Label considered as positive and others are considered negative.
  sample_weight : array-like of shape = [n_samples], optional
      Sample weights.
  drop_intermediate : boolean, optional (default=True)
      Whether to drop some suboptimal thresholds which would not appear
      on a plotted ROC curve. This is useful in order to create lighter
      ROC curves.

  Returns
  -------
  fpr : array, shape = [>2]
      Increasing false positive rates such that element i is the false
      positive rate of predictions with score >= thresholds[i].
  tpr : array, shape = [>2]
      Increasing true positive rates such that element i is the true
      positive rate of predictions with score >= thresholds[i].
  thresholds : array, shape = [n_thresholds]
      Decreasing thresholds on the decision function used to compute
      fpr and tpr. `thresholds[0]` represents no instances being predicted
      and is arbitrarily set to `max(y_score) + 1`.

  Notes
  -----
  Since the thresholds are sorted from low to high values, they
  are reversed upon returning them to ensure they correspond to both ``fpr``
  and ``tpr``, which are sorted in reversed order during their calculation.

  References
  ----------
  .. [1] `Wikipedia entry for the Receiver operating characteristic
          <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

  Examples
  --------
  >>> import numpy as np
  >>> from sklearn import metrics
  >>> y = np.array([1, 1, 2, 2])
  >>> scores = np.array([0.1, 0.4, 0.35, 0.8])
  >>> fpr, tpr, thresholds = metrics.roc_curve(y, scores, pos_label=2)
  >>> fpr
  array([ 0. ,  0.5,  0.5,  1. ])
  >>> tpr
  array([ 0.5,  0.5,  1. ,  1. ])
  >>> thresholds
  array([ 0.8 ,  0.4 ,  0.35,  0.1 ])

  """
  from sklearn.metrics import roc_curve
  return roc_curve(y_true, y_score, pos_label, sample_weight, drop_intermediate)


def prc_curve(y_true, y_probas, pos_label=None, sample_weight=None):
  """Compute precision-recall pairs for different probability thresholds

  Note: this implementation is restricted to the binary classification task.

  The precision is the ratio ``tp / (tp + fp)`` where ``tp`` is the number of
  true positives and ``fp`` the number of false positives. The precision is
  intuitively the ability of the classifier not to label as positive a sample
  that is negative.

  The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
  true positives and ``fn`` the number of false negatives. The recall is
  intuitively the ability of the classifier to find all the positive samples.

  The last precision and recall values are 1. and 0. respectively and do not
  have a corresponding threshold.  This ensures that the graph starts on the
  x axis.

  Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

  Parameters
  ----------
  y_true : array, shape = [n_samples]
      True targets of binary classification in range {-1, 1} or {0, 1}.
  y_probas : array, shape = [n_samples]
      Estimated probabilities or decision function.
  pos_label : int or str, default=None
      The label of the positive class
  sample_weight : array-like of shape = [n_samples], optional
      Sample weights.

  Returns
  -------
  precision : array, shape = [n_thresholds + 1]
      Precision values such that element i is the precision of
      predictions with score >= thresholds[i] and the last element is 1.
  recall : array, shape = [n_thresholds + 1]
      Decreasing recall values such that element i is the recall of
      predictions with score >= thresholds[i] and the last element is 0.
  thresholds : array, shape = [n_thresholds <= len(np.unique(probas_pred))]
      Increasing thresholds on the decision function used to compute
      precision and recall.

  Examples
  --------
  >>> import numpy as np
  >>> from sklearn.metrics import precision_recall_curve
  >>> y_true = np.array([0, 0, 1, 1])
  >>> y_scores = np.array([0.1, 0.4, 0.35, 0.8])
  >>> precision, recall, thresholds = precision_recall_curve(
  ...     y_true, y_scores)
  >>> precision  # doctest: +ELLIPSIS
  array([ 0.66...,  0.5       ,  1.        ,  1.        ])
  >>> recall
  array([ 1. ,  0.5,  0.5,  0. ])
  >>> thresholds
  array([ 0.35,  0.4 ,  0.8 ])

  """
  from sklearn.metrics import precision_recall_curve
  return precision_recall_curve(y_true, y_probas, pos_label, sample_weight)


def det_curve(y_true, y_score, pos_label=None, sample_weight=None):
  """Detection Error Tradeoff
  Compute error rates for different probability thresholds

  @Original implementaion from NIST
  The function is adapted to take input format same as
  NIST original code and `sklearn.metrics`

  Note: this implementation is restricted to the binary classification task.
  (By convention, the more positive the score,
  the more likely is the target hypothesis.)

  Parameters
  ----------
  y_true : array, shape = [n_samples]
      True targets of binary classification in range {-1, 1} or {0, 1}.
  y_score : array, shape = [n_samples]
      Estimated probabilities or decision function.
  pos_label : int, optional (default=None)
      The label of the positive class
  sample_weight : array-like of shape = [n_samples], optional
      Sample weights.

  Returns
  -------
  with `n_samples = n_true_samples + n_false_samples`
  P_fa: array, shape = [n_samples]
      fpr - False Positive rate, or false alarm probabilities
  P_miss : array, shape = [n_samples]
      fnr - False Negative rate, or miss probabilities

  References
  ----------
  .. [1] `Wikipedia entry for Detection error tradeoff
          <https://en.wikipedia.org/wiki/Detection_error_tradeoff>`_
  .. [2] `The DET Curve in Assessment of Detection Task Performance
          <http://www.itl.nist.gov/iad/mig/publications/storage_paper/det.pdf>`_
  .. [3] `2008 NIST Speaker Recognition Evaluation Results
          <http://www.itl.nist.gov/iad/mig/tests/sre/2008/official_results/>`_
  .. [4] `DET-Curve Plotting software for use with MATLAB
          <http://www.itl.nist.gov/iad/mig/tools/DETware_v2.1.targz.htm>`_

  Examples
  --------
  >>> import numpy as np
  >>> from odin import backend as K
  >>> y_true = np.array([0, 0, 1, 1])
  >>> y_scores = np.array([0.1, 0.4, 0.35, 0.8])
  >>> fnr, fpr = K.metrics.det_curve(y_true, y_scores)
  >>> print(fpr)
  array([ 0.5,  0.5,  0. ])
  >>> print(fnr)
  array([ 0. ,  0.5,  0.5])
  >>> print(thresholds)
  array([ 0.35,  0.4 ,  0.8 ])
  """
  # ====== ravel everything in cased of multi-classes ====== #
  y_score = y_score.ravel()
  y_true = np.array(y_true)
  if y_true.ndim >= 2:
    y_true = np.argmax(y_true, axis=-1)
  nb_classes = len(np.lib.arraysetops.unique(y_true))
  # multi-classes
  if nb_classes > 2:
    total_samples = nb_classes * len(y_true)
    indices = np.arange(0, total_samples, nb_classes) + y_true
    y_true = np.zeros(total_samples, dtype=np.int)
    y_true[indices] = 1
  # ====== check weights ====== #
  if sample_weight is not None:
    if len(sample_weight) != len(y_score):
      raise ValueError("Provided `sample_weight` for %d samples, but got "
                       "scores for %d samples." %
                       (len(sample_weight), len(y_score)))
  else:
    sample_weight = np.ones(shape=(len(y_score),), dtype=y_score.dtype)
  # ====== processing ====== #
  if pos_label is not None:
    y_true = (y_true == pos_label).astype(np.int)
  # ====== start ====== #
  sorted_ndx = np.argsort(y_score)
  y_true = y_true[sorted_ndx]
  # sort the weights also, dont forget this
  sample_weight = sample_weight[sorted_ndx]
  tgt_weights = sample_weight * y_true
  imp_weights = sample_weight * (1 - y_true)
  # FNR
  Pmiss = np.cumsum(tgt_weights) / np.sum(tgt_weights)
  # FPR
  Pfa = 1 - np.cumsum(imp_weights) / np.sum(imp_weights)
  return Pfa, Pmiss



from __future__ import absolute_import, division, print_function

import tensorflow as tf
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python import layers as tfl
from tensorflow_probability.python.internal import \
    distribution_util as dist_util
from tensorflow_probability.python.layers import DistributionLambda
from tensorflow_probability.python.layers.distribution_layer import _event_size

from odin.bay.distributions import NegativeBinomialDisp, ZeroInflated

__all__ = [
    'PoissonLayer', 'NegativeBinomialDispLayer', 'NegativeBinomialLayer',
    'ZINegativeBinomialDispLayer', 'ZINegativeBinomialLayer', 'ZIPoissonLayer'
]

PoissonLayer = tfl.IndependentPoisson


class NegativeBinomialLayer(DistributionLambda):
  """An independent NegativeBinomial Keras layer.

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.
  given_log_count : boolean
    is the input representing log count values or the count itself
  dispersion : {'full', 'share', 'single'}
    'full' creates a dispersion value for each individual data point,
    'share' creates a single vector of dispersion for all examples, and
    'single' uses a single value as dispersion for all data points.
    Note: the dispersion in this case is the probability of success.
  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.
  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               given_log_count=True,
               dispersion='full',
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    dispersion = str(dispersion).lower()
    assert dispersion in ('full', 'single', 'share'), \
      "Only support three different dispersion value: 'full', 'single' and " + \
        "'share', but given: %s" % dispersion
    super(NegativeBinomialLayer,
          self).__init__(lambda t: type(self).new(
              t, event_shape, given_log_count, dispersion, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          given_log_count=True,
          dispersion='full',
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'NegativeBinomialLayer',
                                 [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      ndims = output_shape.shape[0]

      total_count_params, logits_params = tf.split(params, 2, axis=-1)
      if dispersion == 'single':
        logits_params = tf.reduce_mean(logits_params)
      elif dispersion == 'share':
        logits_params = tf.reduce_mean(logits_params,
                                       axis=tf.range(0,
                                                     ndims - 1,
                                                     dtype='int32'),
                                       keepdims=True)
      if given_log_count:
        total_count_params = tf.exp(total_count_params, name='total_count')
      return tfd.Independent(
          tfd.NegativeBinomial(total_count=tf.reshape(total_count_params,
                                                      output_shape),
                               logits=tf.reshape(logits_params, output_shape)
                               if dispersion == 'full' else logits_params,
                               validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'NegativeBinomial_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape,
                             name=name or 'NegativeBinomial_params_size')


class NegativeBinomialDispLayer(DistributionLambda):
  """An alternative parameterization of the NegativeBinomial Keras layer.

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.
  given_log_mean : `bool`
    is the input representing log mean values or the count mean itself
  given_log_mean : `bool`
    is the input representing log mean values or the count mean itself
  dispersion : {'full', 'share', 'single'}
    'full' creates a dispersion value for each individual data point,
    'share' creates a single vector of dispersion for all examples, and
    'single' uses a single value as dispersion for all data points.
  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.
  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               given_log_mean=True,
               given_log_disp=True,
               dispersion='full',
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    dispersion = str(dispersion).lower()
    assert dispersion in ('full', 'single', 'share'), \
      "Only support three different dispersion value: 'full', 'single' and " + \
        "'share', but given: %s" % dispersion
    super(NegativeBinomialDispLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, given_log_mean, given_log_disp,
                                 dispersion, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          given_log_mean=True,
          given_log_disp=True,
          dispersion='full',
          validate_args=False,
          name=None):
    """ Create the distribution instance from a `params` vector. """
    with tf.compat.v1.name_scope(name, 'NegativeBinomialDispLayer',
                                 [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      loc_params, disp_params = tf.split(params, 2, axis=-1)
      if dispersion == 'single':
        disp_params = tf.reduce_mean(disp_params)
      elif dispersion == 'share':
        disp_params = tf.reduce_mean(disp_params,
                                     axis=tf.range(0,
                                                   output_shape.shape[0] - 1,
                                                   dtype='int32'),
                                     keepdims=True)
      if given_log_mean:
        loc_params = tf.exp(loc_params, name='loc')
      if given_log_disp:
        disp_params = tf.exp(disp_params, name='disp')
      return tfd.Independent(
          NegativeBinomialDisp(loc=tf.reshape(loc_params, output_shape),
                               disp=tf.reshape(disp_params, output_shape)
                               if dispersion == 'full' else disp_params,
                               validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'NegativeBinomialDisp_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape,
                             name=name or 'NegativeBinomialDisp_params_size')


# ===========================================================================
# Zero inflated
# ===========================================================================
class ZIPoissonLayer(DistributionLambda):
  """A Independent zero-inflated Poisson keras layer
  """

  def __init__(self,
               event_shape=(),
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(ZIPoissonLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params, event_shape=(), validate_args=False, name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'ZIPoissonLayer', [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      (log_rate_params, logits_params) = tf.split(params, 2, axis=-1)
      zip = ZeroInflated(count_distribution=tfd.Poisson(
          log_rate=tf.reshape(log_rate_params, output_shape),
          validate_args=validate_args),
                         logits=tf.reshape(logits_params, output_shape),
                         validate_args=validate_args)
      return tfd.Independent(
          zip,
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name,
                                 'ZeroInflatedNegativeBinomial_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(
          event_shape, name=name or 'ZeroInflatedNegativeBinomial_params_size')


class ZINegativeBinomialLayer(DistributionLambda):
  """A Independent zero-inflated negative binomial keras layer

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.
  given_log_count : boolean
    is the input representing log count values or the count itself
  dispersion : {'full', 'share', 'single'}
    'full' creates a dispersion value for each individual data point,
    'share' creates a single vector of dispersion for all examples, and
    'single' uses a single value as dispersion for all data points.
  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.
  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               given_log_count=True,
               dispersion='full',
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(ZINegativeBinomialLayer,
          self).__init__(lambda t: type(self).new(
              t, event_shape, given_log_count, dispersion, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          given_log_count=True,
          dispersion='full',
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'ZINegativeBinomialLayer',
                                 [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      ndims = output_shape.shape[0]
      (total_count_params, logits_params, rate_params) = tf.split(params,
                                                                  3,
                                                                  axis=-1)
      if dispersion == 'single':
        logits_params = tf.reduce_mean(logits_params)
      elif dispersion == 'share':
        logits_params = tf.reduce_mean(logits_params,
                                       axis=tf.range(0,
                                                     ndims - 1,
                                                     dtype='int32'),
                                       keepdims=True)
      if given_log_count:
        total_count_params = tf.exp(total_count_params, name='total_count')
      nb = tfd.NegativeBinomial(total_count=tf.reshape(total_count_params,
                                                       output_shape),
                                logits=tf.reshape(logits_params, output_shape)
                                if dispersion == 'full' else logits_params,
                                validate_args=validate_args)
      zinb = ZeroInflated(count_distribution=nb,
                          logits=tf.reshape(rate_params, output_shape),
                          validate_args=validate_args)
      return tfd.Independent(
          zinb,
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name,
                                 'ZeroInflatedNegativeBinomial_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 3 * _event_size(
          event_shape, name=name or 'ZeroInflatedNegativeBinomial_params_size')


class ZINegativeBinomialDispLayer(DistributionLambda):
  """A Independent zero-inflated negative binomial (alternative
  parameterization) keras layer

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.
  given_log_mean : boolean
    is the input representing log count values or the count itself
  given_log_disp : boolean
    is the input representing log dispersion values
  dispersion : {'full', 'share', 'single'}
    'full' creates a dispersion value for each individual data point,
    'share' creates a single vector of dispersion for all examples, and
    'single' uses a single value as dispersion for all data points.
  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.
  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               given_log_mean=True,
               given_log_disp=True,
               dispersion='full',
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(ZINegativeBinomialDispLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, given_log_mean, given_log_disp,
                                 dispersion, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          given_log_mean=True,
          given_log_disp=True,
          dispersion='full',
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'ZINegativeBinomialDispLayer',
                                 [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      # splitting the parameters
      (loc_params, disp_params, rate_params) = tf.split(params, 3, axis=-1)
      if dispersion == 'single':
        disp_params = tf.reduce_mean(disp_params)
      elif dispersion == 'share':
        disp_params = tf.reduce_mean(disp_params,
                                     axis=tf.range(0,
                                                   output_shape.shape[0] - 1,
                                                   dtype='int32'),
                                     keepdims=True)
      # as count value, do exp if necessary
      if given_log_mean:
        loc_params = tf.exp(loc_params, name='loc')
      if given_log_disp:
        disp_params = tf.exp(disp_params, name='disp')
      # create the distribution
      nb = NegativeBinomialDisp(loc=tf.reshape(loc_params, output_shape),
                                disp=tf.reshape(disp_params, output_shape)
                                if dispersion == 'full' else disp_params,
                                validate_args=validate_args)
      zinb = ZeroInflated(count_distribution=nb,
                          logits=tf.reshape(rate_params, output_shape),
                          validate_args=validate_args)
      return tfd.Independent(
          zinb,
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'ZINegativeBinomialDisp_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 3 * _event_size(event_shape,
                             name=name or 'ZINegativeBinomialDisp_params_size')



from __future__ import print_function, division, absolute_import

import math
import numpy as np

# ===========================================================================
# Distance measurement
# ===========================================================================
def _LevenshteinDistance(s1, s2):
  ''' Implementation of the wikipedia algorithm, optimized for memory
  Reference: http://rosettacode.org/wiki/Levenshtein_distance#Python
  '''
  if len(s1) > len(s2):
    s1, s2 = s2, s1
  distances = range(len(s1) + 1)
  for index2, char2 in enumerate(s2):
    newDistances = [index2 + 1]
    for index1, char1 in enumerate(s1):
      if char1 == char2:
        newDistances.append(distances[index1])
      else:
        newDistances.append(1 + min((distances[index1],
                                     distances[index1 + 1],
                                     newDistances[-1])))
    distances = newDistances
  return distances[-1]

try:
  # Cython implementation about 100 time faster
  import Levenshtein
  Levenshtein_distance = Levenshtein.distance
except ImportError as e:
  # python implementation
  Levenshtein_distance = _LevenshteinDistance

edit_distance = Levenshtein_distance

def LER(y_true, y_pred, return_mean=True):
  ''' This function calculates the Labelling Error Rate (PER) of the decoded
  networks output sequence (out) and a target sequence (tar) with Levenshtein
  distance and dynamic programming. This is the same algorithm as commonly used
  for calculating the word error rate (WER), or phonemes error rate (PER).

  Parameters
  ----------
  y_true : ndarray (nb_samples, seq_labels)
      true values of sequences
  y_pred : ndarray (nb_samples, seq_labels)
      prediction values of sequences

  Returns
  -------
  return : float
      Labelling error rate
  '''
  if not hasattr(y_true[0], '__len__') or isinstance(y_true[0], str):
    y_true = [y_true]
  if not hasattr(y_pred[0], '__len__') or isinstance(y_pred[0], str):
    y_pred = [y_pred]

  results = []
  for ytrue, ypred in zip(y_true, y_pred):
    results.append(Levenshtein_distance(ytrue, ypred) / len(ytrue))
  if return_mean:
    return np.mean(results)
  return results

# ===========================================================================
# Interpolation
# ===========================================================================
def _power_interp(a, power):
  return np.where(a <= 0.5,
                  np.power(a * 2, power) / 2,
                  np.power((a - 1) * 2, power) / (-2 if power % 2 == 0 else 2) + 1)


def _exp_interp(a, value, power):
  min = np.power(value, -float(power))
  scale = 1 / (1 - min)
  return np.where(a <= 0.5,
      (np.power(value, power * (a * 2 - 1)) - min) * scale / 2,
      (2 - (np.power(value, -power * (a * 2 - 1)) - min) * scale) / 2)


def _expIn_interp(a, value, power):
  min = np.power(value, -float(power))
  scale = 1 / (1 - min)
  return (np.power(value, power * (a - 1)) - min) * scale


def _expOut_interp(a, value, power):
  min = np.power(value, -float(power))
  scale = 1 / (1 - min)
  return 1 - (np.power(value, -power * a) - min) * scale


def _0_1_array(n):
  return np.linspace(start=0., stop=1., num=int(n))


class interp(object):
  """ Original code, libgdx
  https://github.com/libgdx/libgdx/wiki/Interpolation

  Return an array of interpolated values within given range: [0., 1.]
  """
  @staticmethod
  def circle(n):
    a = _0_1_array(n)
    a1 = (a - 1) * 2
    return np.where(a <= 0.5,
                    (1 - np.sqrt(1 - np.power(2 * a, 2.))) / 2,
                    (np.sqrt(1 - np.power(a1, 2.) + 1) / 2))

  @staticmethod
  def pow(n, power):
    return _power_interp(_0_1_array(n), power)

  @staticmethod
  def powIn(n, power):
    return np.power(_0_1_array(n), power)

  @staticmethod
  def powOut(n, power):
    return np.power(_0_1_array(n) - 1, power) * (-1 if power % 2 == 0 else 1) + 1

  @staticmethod
  def exp(n, power):
    return _exp_interp(_0_1_array(n), value=2, power=power)

  @staticmethod
  def expIn(n, power):
    return _expIn_interp(_0_1_array(n), value=2, power=power)

  @staticmethod
  def expOut(n, power):
    return _expOut_interp(_0_1_array(n), value=2, power=power)

  @staticmethod
  def swing(n, scale=1.5):
    scale = scale * 2
    a = _0_1_array(n)
    a1 = (a - 1) * 2
    a2 = a * 2
    return np.where(a <= 0.5,
                    np.power(a2, 2.) * ((scale + 1) * a2 - scale) / 2,
                    np.power(a1, 2.) * ((scale + 1) * a1 + scale) / 2 + 1)

  @staticmethod
  def swingIn(n, scale=2):
    a = _0_1_array(n)
    return a * a * ((scale + 1) * a - scale)

  @staticmethod
  def swingOut(n, scale=2):
    a = _0_1_array(n) - 1.
    return a * a * ((scale + 1) * a + scale) + 1



from __future__ import absolute_import, division, print_function

import os
import pickle
import shutil
import warnings
from collections import OrderedDict, defaultdict
from enum import Enum
from numbers import Number

import numba as nb
import numpy as np
from scipy.io import wavfile

from odin import fuel as F
from odin import visual as V
from odin.stats import freqcount, sampling_iter
from odin.utils import (Progbar, args_parse, cache_disk, catch_warnings_error,
                        catch_warnings_ignore, crypto, ctext, get_exppath,
                        get_logpath, get_module_from_path, get_script_name,
                        get_script_path, mpi, select_path)


# ===========================================================================
# Configuration
# ===========================================================================
class Config(object):
  # ====== Acoustic features ====== #
  FRAME_LENGTH = 0.025
  STEP_LENGTH = 0.01
  SAMPLE_RATE = 8000
  WINDOW = 'hamm'
  NFFT = 512
  # Random seed for reproducibility
  SUPER_SEED = 87654321

class SystemStates(Enum):
  """ SystemStates """
  UNKNOWN = 0
  EXTRACT_FEATURES = 1
  TRAINING = 2
  SCORING = 3

# ===========================================================================
# General arguments for all experiments
# ===========================================================================
_args = args_parse(descriptions=[
    ('recipe', 'recipe is the name of acoustic Dataset defined in feature_recipes.py', None),
    ('-feat', 'specific name for the acoustic features, extracted from the given recipe', None, ''),
    ('-aug', 'augmentation dataset: musan, rirs; could be multiple dataset '
             'for training: "musan,rirs"', None, 'None'),
    ('-ncpu', 'number of CPU to be used, if <= 0, auto-select', None, 0),
    # for scoring
    ('-sys', 'name of the system for scoring: xvec, ivec, e2e ...', None, 'xvec'),
    ('-sysid', 'when a system is saved multiple checkpoint (e.g. sys.0.ai)', None, '-1'),
    ('-score', 'name of dataset for scoring, multiple dataset split by ","', None, 'sre18dev,sre18eval'),
    ('-backend', 'list of dataset for training the backend: '
                 'PLDA, SVM or Cosine', None, 'sre04,sre05,sre06,sre08,sre10,mx6'),
    ('-lda', 'if > 0, running LDA before training the backend '
             'with given number of components', None, 0),
    ('-plda', 'number of PLDA components, must be > 0 ', None, 150),
    ('--mll', 'pre-fitting maximum likelihood before training PLDA', None, False),
    ('--showllk', 'show LLK during training of PLDA, this will slow thing down', None, False),
    # for training
    ('-downsample', 'absolute number of files used for training', None, 0),
    ('-exclude', 'list of excluded dataset not for training,'
                 'multiple dataset split by ","', None, ''),
    # for ivector
    ('-nmix', 'for i-vector training, number of Gaussian components', None, 2048),
    ('-tdim', 'for i-vector training, number of latent dimension for i-vector', None, 600),
    # for DNN
    ('-utt', 'maximum length of sequence for training', None, 3),
    ('-seq', 'sequencing mode for training data, cut or pad', None, 'cut'),
    ('-batch', 'batch size, for training DNN, kaldi use 64, we use 128', None, 128),
    ('-epoch', 'number of epoch, for training DNN, kaldi only 3 epochs', None, 12),
    ('-clip', 'The maximum change in parameters allowed per minibatch, '
              'measured in Euclidean norm over the entire model (change '
              'will be clipped to this value), kaldi use 2.0', None, 2.0),
    ('-lr', 'learning rate for Adam, kaldi use 0.001 by default,'
            ' we use 0.01', None, 0.01),
    # others
    ('-mindur', 'for filtering utterances, minimum duration of utterance '
                'for training (in second)', None, 1),
    ('-minutt', 'for filtering utterances, minimum number of utterance of '
                'each speaker for training', None, 3),
    ('--override', 'override previous experiments', None, False),
    ('--debug', 'enable debugging', None, False),
])
IS_DEBUGGING = bool(_args.debug)
IS_OVERRIDE = bool(_args.override)
MINIMUM_UTT_DURATION = int(_args.mindur) # in seconds
assert MINIMUM_UTT_DURATION > 0, "Minimum utterances duration must be greater than 0"
MINIMUM_UTT_PER_SPEAKERS = int(_args.minutt) # number of utterances
# this variable determine which state is running
CURRENT_STATE = SystemStates.UNKNOWN
# ====== Features extraction ====== #
FEATURE_RECIPE = str(_args.recipe)
FEATURE_NAME = FEATURE_RECIPE.split('_')[0] if len(str(_args.feat)) == 0 else str(_args.feat)
AUGMENTATION_NAME = _args.aug
TRAINING_DATASET = ['mx6', 'voxceleb1', 'voxceleb2', 'swb', 'fisher',
                    'sre04', 'sre05', 'sre06', 'sre08', 'sre10']
# ====== DNN ====== #
BATCH_SIZE = int(_args.batch)
EPOCH = int(_args.epoch)
LEARNING_RATE = float(_args.lr)
GRADIENT_CLIPPING = float(_args.clip)
# ====== searching for the appropriate system ====== #
SCORE_SYSTEM_NAME = _args.sys
SCORE_SYSTEM_ID = int(_args.sysid)
N_LDA = int(_args.lda)
N_PLDA = int(_args.plda)
assert N_PLDA > 0, "Number of PLDA components must > 0, but given: %d" % N_PLDA
PLDA_MAXIMUM_LIKELIHOOD = bool(_args.mll)
PLDA_SHOW_LLK = bool(_args.showllk)
# ====== system ====== #
NCPU = min(18, mpi.cpu_count() - 2) if _args.ncpu <= 0 else int(_args.ncpu)
# ====== helper for checking the requirement ====== #
def _check_feature_extraction_requirement():
  # check requirement for feature extraction
  from shutil import which
  if which('sox') is None:
    raise RuntimeError("`sox` was not installed")
  if which('sph2pipe') is None:
    raise RuntimeError("`sph2pipe` was not installed")
  if which('ffmpeg') is None:
    raise RuntimeError("`ffmpeg` was not installed")

def _check_recipe_name_for_extraction():
  # check the requirement of recipe name for feature extraction
  if '_' in FEATURE_RECIPE:
    raise ValueError("'_' can appear in recipe name which is: '%s'" % FEATURE_RECIPE)
# ====== check the running script to determine the current running states ====== #
_script_name = get_script_name()
if _script_name in ('speech_augmentation', 'speech_features_extraction'):
  CURRENT_STATE = SystemStates.EXTRACT_FEATURES
  _check_feature_extraction_requirement()
  _check_recipe_name_for_extraction()
elif _script_name in ('train_xvec', 'train_ivec', 'train_tvec',
                      'train_evec', 'analyze', 'analyze_data'):
  CURRENT_STATE = SystemStates.TRAINING
elif _script_name in ('make_score'):
  CURRENT_STATE = SystemStates.SCORING
  _check_feature_extraction_requirement()
else:
  raise RuntimeError("Unknown states for current running script: %s/%s" %
    (get_script_path(), get_script_name()))
# some fancy log of current state
print(ctext('====================================', 'red'))
print(ctext("System state:", 'cyan'), ctext(CURRENT_STATE, 'yellow'))
print(ctext('====================================', 'red'))
# ===========================================================================
# FILE LIST PATH
# ===========================================================================
# ====== basic directories ====== #
EXP_DIR = get_exppath('sre', override=False)
# this folder store extracted vectors for training backend and extracting scores
VECTORS_DIR = os.path.join(EXP_DIR, 'vectors')
if not os.path.exists(VECTORS_DIR):
  os.mkdir(VECTORS_DIR)
# this folder store the results
RESULT_DIR = os.path.join(EXP_DIR, 'results')
if not os.path.exists(RESULT_DIR):
  os.mkdir(RESULT_DIR)
# this folder store the analysis
ANALYSIS_DIR = os.path.join(EXP_DIR, 'analysis')
if not os.path.exists(ANALYSIS_DIR):
  os.mkdir(ANALYSIS_DIR)
# ====== raw data ====== #
PATH_BASE = select_path(
    '/media/data2/SRE_DATA',
    '/mnt/sda1/SRE_DATA',
    '/mnt/sdb1/SRE_DATA',
default='')
# path to directory contain following folders:
##############
#   * fisher
#   * mx6
#   * sre04
#   * sre05
#   * sre06
#   * sre08
#   * sre10
#   * swb
#   * voxceleb1
#   * voxceleb2
###############
#   * musan
#   * rirs
###############
#   * sre18dev
#   * sre18eval
PATH_RAW_DATA = {
    'mx6': PATH_BASE,
    'voxceleb1': PATH_BASE,
    'voxceleb2': PATH_BASE,
    'swb': PATH_BASE,
    'fisher': PATH_BASE,
    'sre04': os.path.join(PATH_BASE, 'NIST1996_2008/SRE02_SRE06'),
    'sre05': os.path.join(PATH_BASE, 'NIST1996_2008/SRE96_SRE05'),
    'sre06': os.path.join(PATH_BASE, 'NIST1996_2008/SRE02_SRE06'),
    'sre08': PATH_BASE,
    'sre10': PATH_BASE,
    'sre18dev': PATH_BASE,
    'sre18eval': PATH_BASE,
    # noise datasets
    'musan': PATH_BASE,
    'rirs': PATH_BASE,
}
# all features will be stored here
OUTPUT_DIR = select_path(
    '/home/trung/data',
    '/media/data1',
    '/mnt/sda1'
)
PATH_ACOUSTIC_FEATURES = os.path.join(OUTPUT_DIR, "SRE_FEAT")
if not os.path.exists(PATH_ACOUSTIC_FEATURES):
  os.mkdir(PATH_ACOUSTIC_FEATURES)
# ===========================================================================
# Load the file list
# ===========================================================================
sre_file_list = F.load_sre_list()
print('README at:', ctext(sre_file_list['README.txt'], 'cyan'))
sre_file_list = {k: v
                 for k, v in sre_file_list.items()
                 if isinstance(v, np.ndarray)}
print("Original dataset:")
for k, v in sorted(sre_file_list.items(), key=lambda x: x[0]):
  print(' ', ctext('%-18s' % k, 'yellow'), ':',
    ctext(v.shape, 'cyan'))
# ===========================================================================
# Validate scoring dataset
# ===========================================================================
def validate_scoring_dataset(in_path_raw, score_dataset, file_must_exist=True):
  all_files = {}
  for dsname in score_dataset:
    if dsname not in sre_file_list:
      raise ValueError("Cannot find dataset with name: '%s' in the file list" % dsname)
    if dsname not in in_path_raw:
      raise ValueError("Cannot find dataset with name: '%s' in provided path" % dsname)
    base_path = in_path_raw[dsname]

    ds = []
    for row in sre_file_list[dsname]:
      path = os.path.join(base_path, row[0])
      # every file must exist
      if bool(file_must_exist) and not os.path.exists(path):
        raise RuntimeError("File not exist at path: %s" % path)
      ds.append([path] + row[1:4].tolist() + [dsname])
    all_files[dsname] = np.array(ds)
  # Header:
  #  0       1      2        3           4
  # path, channel, name, something, dataset_name
  return all_files

# ====== check dataset for scoring ====== #
if CURRENT_STATE == SystemStates.SCORING:
  assert len(_args.score) > 0, \
  "No dataset are provided for scoring, specify '-score' option"

  # for scoring
  SCORING_DATASETS = validate_scoring_dataset(
      in_path_raw=PATH_RAW_DATA,
      score_dataset=str(_args.score).strip().split(','))
  print("Processed scoring dataset:")
  for dsname, dsarray in sorted(SCORING_DATASETS.items(),
                                key=lambda x: x[0]):
    print('  ', ctext('%-10s' % dsname, 'yellow'), ':',
          '%s' % ctext(dsarray.shape, 'cyan'))

  # for training the backend
  BACKEND_DATASETS = validate_scoring_dataset(
      in_path_raw=PATH_RAW_DATA,
      score_dataset=str(_args.backend).strip().split(','),
      file_must_exist=False)
  assert len(BACKEND_DATASETS) > 0, \
  "Datasets for training the backend must be provided"
  print("Processed backend dataset:")
  for dsname, dsarray in sorted(BACKEND_DATASETS.items(),
                                key=lambda x: x[0]):
    print('  ', ctext('%-10s' % dsname, 'yellow'), ':',
          '%s' % ctext(dsarray.shape, 'cyan'))
# ===========================================================================
# Validating the Noise dataset for augmentation
# ===========================================================================
@cache_disk
def validating_noise_data(in_path_raw):
  # preparing
  noise_dataset = ['musan', 'rirs']
  all_files = defaultdict(list)
  n_files = sum(len(sre_file_list[i])
                for i in noise_dataset
                if i in sre_file_list)
  n_non_exist = 0
  n_exist = 0
  prog = Progbar(target=n_files, print_summary=True,
                 name="Validating noise dataset")
  prog.set_summarizer(key='#Non-exist', fn=lambda x: x[-1])
  prog.set_summarizer(key='#Exist', fn=lambda x: x[-1])
  # check all dataset
  for ds_name in noise_dataset:
    if ds_name not in sre_file_list:
      continue
    if ds_name not in in_path_raw:
      continue
    base_path = in_path_raw[ds_name]
    base_ds = all_files[ds_name]
    # start validating
    for row in sre_file_list[ds_name]:
      # check file
      path, channel, name, noise_type, duration = row[:5]
      path = os.path.join(base_path, path)
      if os.path.exists(path):
        base_ds.append([path, channel, name, noise_type, duration])
        n_exist += 1
      else:
        n_non_exist += 1
      # update progress
      prog['ds'] = ds_name
      prog['#Exist'] = n_exist
      prog['#Non-exist'] = n_non_exist
      prog.add(1)
  # ====== return ====== #
  # Header:
  #  0       1      2         3         4
  # path, channel, name, noise_type, duration
  return {key: np.array(sorted(val, key=lambda x: x[0]))
          for key, val in all_files.items()}
# ==================== run the validation ==================== #
if CURRENT_STATE == SystemStates.EXTRACT_FEATURES:
  ALL_NOISE = validating_noise_data(
      in_path_raw=PATH_RAW_DATA)
  print("Processed noise data:")
  for ds_name, noise_list in ALL_NOISE.items():
    print(" ", ctext(ds_name, 'yellow'), ':', noise_list.shape)
    if len(noise_list) == 0:
      continue
    for name, count in sorted(freqcount(noise_list[:, 3]).items(),
                              key=lambda x: x[0]):
      print('  ', ctext('%-10s' % name, 'yellow'), ':',
            '%s(files)' % ctext('%-6d' % count, 'cyan'))
# ===========================================================================
# Validating the file list of training data
# ===========================================================================
@cache_disk
def validating_training_data(in_path_raw, training_dataset):
  file_list = {ds: sre_file_list[ds]
               for ds in training_dataset
               if ds in sre_file_list}
  # ====== meta info ====== #
  all_files = []
  non_exist_files = []
  extension_count = defaultdict(int)
  total_data = sum(v.shape[0]
                   for k, v in file_list.items()
                   if k not in('musan', 'rirs'))
  # ====== progress ====== #
  prog = Progbar(target=total_data,
                 print_summary=True, print_report=True,
                 name="Preprocessing File List")
  prog.set_summarizer('#Files', fn=lambda x: x[-1])
  prog.set_summarizer('#Non-exist', fn=lambda x: x[-1])
  # ====== iterating ====== #
  for ds_name, data in sorted(file_list.items(),
                              key=lambda x: x[0]):
    if ds_name in ('musan', 'rirs'):
      continue
    for row in data:
      path, channel, name, spkid = row[:4]
      assert channel in ('0', '1')
      # check path provided
      if ds_name in in_path_raw:
        path = os.path.join(in_path_raw[ds_name], path)
      # create new row
      start_time = '-'
      end_time = '-'
      if ds_name == 'mx6':
        start_time, end_time = row[-2:]
      new_row = [path, channel, name,
                 ds_name + '_' + spkid, ds_name,
                 start_time, end_time]
      # check file exist
      if os.path.exists(path):
        all_files.append(new_row)
      else:
        non_exist_files.append(new_row)
      # extension
      ext = os.path.splitext(path)[-1]
      extension_count[ext + '-' + ds_name] += 1
      # update progress
      prog['Dataset'] = ds_name
      prog['#Files'] = len(all_files)
      prog['#Non-exist'] = len(non_exist_files)
      prog.add(1)
  # final results
  all_files = np.array(all_files)
  if len(all_files) == 0:
    return all_files, np.array(non_exist_files), extension_count
  # ====== check no duplicated name ====== #
  n_files = len(all_files)
  n_unique_files = len(np.unique(all_files[:, 2]))
  assert n_files == n_unique_files, \
  'Found duplicated name: %d != %d' % (n_files, n_unique_files)
  # ====== check no duplicated speaker ====== #
  n_spk = sum(len(np.unique(dat[:, 3]))
              for name, dat in file_list.items()
              if name not in ('musan', 'rirs'))
  n_unique_spk = len(np.unique(all_files[:, 3]))
  assert n_spk == n_unique_spk, \
  'Found duplicated speakers: %d != %d' % (n_spk, n_unique_spk)
  # ====== return ====== #
  # Header:
  #  0       1      2      3       4          5         6
  # path, channel, name, spkid, dataset, start_time, end_time
  return all_files, np.array(non_exist_files), extension_count
# ==================== run the validation process ==================== #
if CURRENT_STATE == SystemStates.EXTRACT_FEATURES:
  (ALL_FILES, NON_EXIST_FILES, ext_count) = validating_training_data(
      in_path_raw=PATH_RAW_DATA,
      training_dataset=TRAINING_DATASET
  )
  if len(ALL_FILES) == 0:
    raise RuntimeError("No files found for feature extraction")

  # list of all dataset
  ALL_DATASET = sorted(np.unique(ALL_FILES[:, 4]))
  print("All extensions:")
  for name, val in sorted(ext_count.items(), key=lambda x: x[0]):
    print('  ', '%-16s' % name, ':', ctext('%-6d' % val, 'cyan'), '(files)')
  print("#Speakers:", ctext(len(np.unique(ALL_FILES[:, 3])), 'cyan'))

  # map Dataset_name -> speaker_ID
  DS_SPK = defaultdict(list)
  for row in ALL_FILES:
    DS_SPK[row[4]].append(row[3])
  DS_SPK = {k: sorted(set(v))
            for k, v in DS_SPK.items()}

  print("Processed datasets:")
  for name, count in sorted(freqcount(ALL_FILES[:, 4]).items(),
                            key=lambda x: x[0]):
    print('  ', ctext('%-10s' % name, 'yellow'), ':',
          '%s(files)' % ctext('%-6d' % count, 'cyan'),
          '%s(spk)' % ctext('%-4d' % len(DS_SPK[name]), 'cyan'))
# ===========================================================================
# PATH HELPER
# ===========================================================================
def get_model_path(system_name, logging=True):
  """
  Parameters
  ----------
  args_name : list of string
    list of name for parsed argument, taken into account for creating
    model name

  Return
  ------
  exp_dir, model_path, log_path
  """
  args_name = []
  if system_name == 'xvec':
    args_name += ['utt', 'seq']
  elif system_name == 'ivec':
    args_name += ['nmix', 'tdim']
  else:
    raise ValueError("No support for system with name: %s" % system_name)
  args_name += ['mindur', 'minutt']
  # ====== base system and feature identity ====== #
  name = str(system_name).lower()
  name += '_' + FEATURE_RECIPE.replace('_', '')
  name += '.' + FEATURE_NAME
  # ====== concat the attributes ====== #
  attributes = []
  for i in [str(i) for i in args_name]:
    attributes.append(str(getattr(_args, i)))
  attributes = '_'.join(attributes)
  name += '.' + attributes
  # ====== check the exclude dataset ====== #
  excluded_dataset = str(_args.exclude).strip()
  if len(excluded_dataset) > 0:
    dataset_str = []
    for excluded in sorted(set(excluded_dataset.split(','))):
      assert excluded in sre_file_list or excluded == 'noise', \
      "Unknown excluded dataset with name: '%s'" % excluded
      dataset_str.append(excluded)
    dataset_str = '_'.join(dataset_str)
    name += '.' + dataset_str
  # ====== check save_path ====== #
  save_path = os.path.join(EXP_DIR, name)
  if os.path.exists(save_path) and IS_OVERRIDE:
    print("Override path:", ctext(save_path, 'yellow'))
    shutil.rmtree(save_path)
  if not os.path.exists(save_path):
    os.mkdir(save_path)
  # ====== return path ====== #
  log_path = get_logpath(name='log.txt', increasing=True,
                         odin_base=False, root=save_path)
  model_path = os.path.join(save_path, 'model.ai')
  if bool(logging):
    print("Model path:", ctext(model_path, 'cyan'))
    print("Log path:", ctext(log_path, 'cyan'))
  return save_path, model_path, log_path
# ===========================================================================
# Data helper
# ===========================================================================
def prepare_dnn_feeder_recipe(name2label=None, n_speakers=None,
                              utt_length=None, seq_mode=None):
  if utt_length is None:
    utt_length = float(_args.utt)
  if seq_mode is None:
    seq_mode = str(_args.seq).strip().lower()
  frame_length = int(utt_length / Config.STEP_LENGTH)

  if seq_mode == 'cut':
    seq_train = 'cut'
    seq_score = 'mix'
  elif seq_mode == 'pad':
    seq_train = 'pad'
    seq_score = 'pad'
  else:
    raise ValueError("Only support 'cut' or 'pad' sequencing mode")

  recipes = [
      F.recipes.Sequencing(frame_length=frame_length,
                           step_length=frame_length,
                           end=seq_score if CURRENT_STATE == SystemStates.SCORING
                           else seq_train,
                           pad_value=0, pad_mode='post',
                           data_idx=0),
  ]
  if name2label is not None and n_speakers is not None:
    recipes += [
        F.recipes.Name2Label(lambda name:name2label[name],
                             ref_idx=0),
        F.recipes.LabelOneHot(nb_classes=n_speakers, data_idx=1)
    ]
  elif (name2label is not None and n_speakers is None) or\
  (name2label is None and n_speakers is not None):
    raise RuntimeError("name2label and n_speakers must both be None, or not-None")
  return recipes

def filter_utterances(X, indices, spkid,
                      min_dur=None, min_utt=None,
                      remove_min_length=True, remove_min_uttspk=True,
                      n_speakers=None, ncpu=None, save_path=None,
                      title=''):
  """
  X : 2-D matrix
    input features

  indices : Mapping
    utterance_name -> (start, end) in `X`

  spkid : Mapping
    utterance_name -> speaker_id

  remove_min_length : bool (default: True)
    if True, remove all files shorter than MINIMUM_UTT_DURATION

  remove_min_uttspk : bool (default: True)
    if True, remove all speakers with lower amount of utterances than
    MINIMUM_UTT_PER_SPEAKERS

  n_speakers : {None, int} (default: None)
    if given, downsample the dataset by given number of speakers

  save_path : {None, str} (default: None)
    if given, pickle all filtered files to disk

  """
  if min_dur is None:
    min_dur = MINIMUM_UTT_DURATION
  if min_utt is None:
    min_utt = MINIMUM_UTT_PER_SPEAKERS

  minimum_amount_of_frames = min_dur / Config.STEP_LENGTH
  save_data = {}

  prog = Progbar(target=len(indices),
                 print_report=True, print_summary=True,
                 name='Filtering broken utterances: %s' % title)
  prog.set_summarizer('zero-length', fn=lambda x: x[-1])
  prog.set_summarizer('min-frames', fn=lambda x: x[-1])
  prog.set_summarizer('zero-var', fn=lambda x: x[-1])
  prog.set_summarizer('small-var', fn=lambda x: x[-1])
  prog.set_summarizer('overflow', fn=lambda x: x[-1])

  # ====== mpi function for checking ====== #
  @nb.jit(nopython=True, nogil=True)
  def _fast_mean_var_ax0(z):
    # using this function for calculating mean and variance
    # can double the speed but cannot check overflow,
    # only accept float32 or float64 input
    s1 = np.zeros(shape=(z.shape[1],), dtype=z.dtype)
    s2 = np.zeros(shape=(z.shape[1],), dtype=z.dtype)
    for i in range(z.shape[0]):
      s1 += z[i]
      s2 += np.power(z[i], 2)
    mean = s1 / z.shape[0]
    var = s2 / z.shape[0] - np.power(mean, 2)
    return mean, var

  def _mpi_func(jobs):
    for name, (start, end) in jobs:
      y = X[start:end]
      # flags
      is_zero_len = False
      is_zero_var = False
      is_small_var = False
      is_min_frames = False
      is_overflow = False
      # checking length
      if y.shape[0] == 0:
        is_zero_len = True
      elif y.shape[0] < minimum_amount_of_frames:
        is_min_frames = True
      # checking statistics
      else:
        with catch_warnings_error(RuntimeWarning):
          try:
            # mean = np.mean(y, axis=-1)
            var = np.var(y, axis=-1)
            # min_val = np.min(y, axis=-1)
            # max_val = np.max(y, axis=-1)
          # numerical unstable
          except RuntimeWarning as w:
            if 'overflow encountered' in str(w):
              is_overflow = True
            else:
              print(name, ':', w)
          # process with more numerical filtering
          else:
            if np.any(np.isclose(var, 0)):
              is_zero_var = True
            # very heuristic and aggressive here
            # filter-out anything with ~16.67% of low-var
            # this could remove 1/3 of the original data
            if np.sum(var < 0.01) > (len(y) / 6):
              is_small_var = True
      # return the flags
      yield (name, is_zero_len, is_min_frames,
             is_zero_var, is_small_var,
             is_overflow)
  # ====== running the multiprocessing filter ====== #
  zero_len_files = {}
  min_frame_files = {}
  zero_var_files = {}
  small_var_files = {}
  overflow_files = {}
  for res in mpi.MPI(jobs=sorted(indices.items(),
                                 key=lambda x: x[1][0]),
                     func=_mpi_func,
                     ncpu=NCPU if ncpu is None else int(ncpu),
                     batch=250):
    name = res[0]
    if res[1]: zero_len_files[name] = 1
    if res[2]: min_frame_files[name] = 1
    if res[3]: zero_var_files[name] = 1
    if res[4]: small_var_files[name] = 1
    if res[5]: overflow_files[name] = 1
    # update progress
    prog['name'] = name[:48]
    prog['zero-length'] = len(zero_len_files)
    prog['min-frames'] = len(min_frame_files)
    prog['zero-var'] = len(zero_var_files)
    prog['small-var'] = len(small_var_files)
    prog['overflow'] = len(overflow_files)
    prog.add(1)
  # ====== remove broken files ====== #
  if not bool(remove_min_length):
    min_frame_files = {}
  new_indices = {name: (start, end)
                 for name, (start, end) in indices.items()
                 if name not in zero_len_files and
                 name not in min_frame_files and
                 name not in zero_var_files and
                 name not in small_var_files and
                 name not in overflow_files}
  print("Filtered #utterances: %s/%s (files)" %
    (ctext(len(indices) - len(new_indices), 'lightcyan'),
     ctext(len(indices), 'cyan')))
  indices = new_indices
  # ====== store save data ====== #
  save_data['zero_len'] = zero_len_files
  save_data['min_dur'] = min_frame_files
  save_data['zero_var'] = zero_var_files
  save_data['small_var'] = small_var_files
  save_data['overflow'] = overflow_files
  # ====== filter-out by number of utt-per-speaker ====== #
  if bool(remove_min_uttspk):
    spk2utt = defaultdict(list)
    for name in indices.keys():
      spk2utt[spkid[name]].append(name)

    n_utt_removed = 0
    n_spk_removed = 0
    removed_utt = []
    keep_utt = []
    for spk, utt in spk2utt.items():
      if len(utt) < min_utt:
        n_utt_removed += len(utt)
        n_spk_removed += 1
        removed_utt += utt
      else:
        keep_utt += utt

    removed_utt = set(removed_utt)
    keep_utt = set(keep_utt)
    save_data['min_utt'] = removed_utt

    print("Removed min-utt/spk:  %s/%s(utt)  %s/%s(spk)" % (
        ctext(n_utt_removed, 'lightcyan'), ctext(len(indices), 'cyan'),
        ctext(n_spk_removed, 'lightcyan'), ctext(len(spk2utt), 'cyan')
    ))
    assert len(indices) == n_utt_removed + len(keep_utt), "Not possible!"

    indices = {name: (start, end)
               for name, (start, end) in indices.items()
               if name in keep_utt}
  # ====== sample by number of speakers ====== #
  if isinstance(n_speakers, Number) and n_speakers > 0:
    spk2utt = defaultdict(list)
    for name, (start, end) in indices.items():
      spk2utt[spkid[name]].append((name, (start, end)))

    n_org_spk = len(spk2utt)
    n_org_ids = len(indices)
    # only need down-sampling with smaller number of speaker
    if n_speakers < n_org_spk:
      rand = np.random.RandomState(seed=Config.SUPER_SEED)
      tmp = list(spk2utt.keys())
      rand.shuffle(tmp)
      sampled_spk = tmp[:n_speakers]

      indices = []
      for spk in sampled_spk:
        indices += spk2utt[spk]
      indices = dict(indices)
    else:
      sampled_spk = spk2utt
    # print some log
    print("Selected: %s/%s(spk) which have %s/%s(utt)" % (
        ctext(len(sampled_spk), 'lightcyan'), ctext(n_org_spk, 'cyan'),
        ctext(len(indices), 'lightcyan'), ctext(n_org_ids, 'cyan')
    ))
  # ====== return the new indices ====== #
  if save_path is not None:
    try:
      with open(save_path, 'wb') as save_file:
        pickle.dump(save_data, save_file)
    except Exception as e:
      print("Cannot save filtering data to path: '%s', error: '%s'" %
        (save_path, str(e)))
  return indices

def prepare_dnn_data(save_dir, feat_name=None,
                     utt_length=None, seq_mode=None,
                     min_dur=None, min_utt=None,
                     exclude=None, train_proportion=None,
                     return_dataset=False):
  assert os.path.isdir(save_dir), \
      "Path to '%s' is not a directory" % save_dir
  if feat_name is None:
    feat_name = FEATURE_NAME
  if utt_length is None:
    utt_length = int(_args.utt)
  if seq_mode is None:
    seq_mode = str(_args.seq).strip().lower()
  if min_dur is None:
    min_dur = MINIMUM_UTT_DURATION
  if min_utt is None:
    min_utt = MINIMUM_UTT_PER_SPEAKERS
  if exclude is None:
    exclude = str(_args.exclude).strip()
  print("Minimum duration: %s(s)" % ctext(min_dur, 'cyan'))
  print("Minimum utt/spk : %s(utt)" % ctext(min_utt, 'cyan'))
  # ******************** prepare dataset ******************** #
  path = os.path.join(PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE)
  assert os.path.exists(path), "Cannot find acoustic dataset at path: %s" % path
  ds = F.Dataset(path=path, read_only=True)
  rand = np.random.RandomState(seed=Config.SUPER_SEED)
  # ====== find the right feature ====== #
  assert feat_name in ds, "Cannot find feature with name: %s" % feat_name
  X = ds[feat_name]
  ids_name = 'indices_%s' % feat_name
  assert ids_name in ds, "Cannot find indices with name: %s" % ids_name
  # ====== basic path ====== #
  path_filtered_data = os.path.join(save_dir, 'filtered_files.pkl')
  path_train_files = os.path.join(save_dir, 'train_files.pkl')
  path_speaker_info = os.path.join(save_dir, 'speaker_info.pkl')
  # ******************** cannot find cached data ******************** #
  if any(not os.path.exists(p) for p in [path_filtered_data,
                                         path_train_files,
                                         path_speaker_info]):
    # ====== exclude some dataset ====== #
    if len(exclude) > 0:
      exclude_dataset = {i: 1 for i in exclude.split(',')}
      print("* Excluded dataset:", ctext(exclude_dataset, 'cyan'))
      indices = {name: (start, end)
                 for name, (start, end) in ds[ids_name].items()
                 if ds['dsname'][name] not in exclude_dataset}
      # special case exclude all the noise data
      if 'noise' in exclude_dataset:
        indices = {name: (start, end)
                   for name, (start, end) in indices.items()
                   if '/' not in name}
    else:
      indices = {i: j for i, j in ds[ids_name].items()}
    # ====== down-sampling if necessary ====== #
    if _args.downsample > 1000:
      dataset2name = defaultdict(list)
      # ordering the indices so we sample the same set every time
      for name in sorted(indices.keys()):
        dataset2name[ds['dsname'][name]].append(name)
      n_total_files = len(indices)
      n_sample_files = int(_args.downsample)
      # get the percentage of each dataset
      dataset2per = {i: len(j) / n_total_files
                     for i, j in dataset2name.items()}
      # sampling based on percentage
      _ = {}
      for dsname, flist in dataset2name.items():
        rand.shuffle(flist)
        n_dataset_files = int(dataset2per[dsname] * n_sample_files)
        _.update({i: indices[i]
                  for i in flist[:n_dataset_files]})
      indices = _
    # ====== * filter out "bad" sample ====== #
    indices = filter_utterances(X=X, indices=indices, spkid=ds['spkid'],
                                min_utt=min_utt, min_dur=min_dur,
                                remove_min_length=True,
                                remove_min_uttspk=True,
                                n_speakers=None, ncpu=None,
                                save_path=path_filtered_data)
    # ====== all training file name ====== #
    # modify here to train full dataset
    all_name = sorted(indices.keys())
    rand.shuffle(all_name); rand.shuffle(all_name)
    n_files = len(all_name)
    print("#Files:", ctext(n_files, 'cyan'))
    # ====== speaker mapping ====== #
    name2spk = {name: ds['spkid'][name]
                for name in all_name}
    all_speakers = sorted(set(name2spk.values()))
    spk2label = {spk: i
                 for i, spk in enumerate(all_speakers)}
    name2label = {name: spk2label[spk]
                  for name, spk in name2spk.items()}
    assert len(name2label) == len(all_name)
    print("#Speakers:", ctext(len(all_speakers), 'cyan'))
    # ====== stratify sampling based on speaker ====== #
    valid_name = []
    # create speakers' cluster
    label2name = defaultdict(list)
    for name, label in sorted(name2label.items(),
                              key=lambda x: x[0]):
      label2name[label].append(name)
    # for each speaker with >= 3 utterance
    for label, name_list in sorted(label2name.items(),
                                   key=lambda x: x[0]):
      if len(name_list) < 3:
        continue
      n = max(1, int(0.05 * len(name_list))) # 5% for validation
      valid_name += rand.choice(a=name_list, size=n, replace=False).tolist()
    # train list is the rest
    _ = set(valid_name)
    train_name = [i for i in all_name if i not in _]
    # ====== split training and validation ====== #
    train_indices = {name: indices[name] for name in train_name}
    valid_indices = {name: indices[name] for name in valid_name}
    # ====== save cached data ====== #
    with open(path_train_files, 'wb') as fout:
      pickle.dump({'train': train_indices, 'valid': valid_indices},
                  fout)
    with open(path_speaker_info, 'wb') as fout:
      pickle.dump({'all_speakers': all_speakers,
                   'name2label': name2label,
                   'spk2label': spk2label},
                  fout)
  # ******************** load cached data ******************** #
  else:
    with open(path_train_files, 'rb') as fin:
      obj = pickle.load(fin)
      train_indices = obj['train']
      valid_indices = obj['valid']
    with open(path_speaker_info, 'rb') as fin:
      obj = pickle.load(fin)
      all_speakers = obj['all_speakers']
      name2label = obj['name2label']
      spk2label = obj['spk2label']

  # ******************** print log ******************** #
  def summary_indices(ids):
    datasets = defaultdict(int)
    speakers = defaultdict(list)
    text = ''
    for name in sorted(ids.keys()):
      text += name + str(ids[name])
      dsname = ds['dsname'][name]
      datasets[dsname] += 1
      speakers[dsname].append(ds['spkid'][name])
    for dsname in sorted(datasets.keys()):
      print('  %-18s: %s(utt) %s(spk)' % (
          dsname,
          ctext('%6d' % datasets[dsname], 'cyan'),
          ctext(len(set(speakers[dsname])), 'cyan')))
    print('  MD5 checksum:', ctext(crypto.md5_checksum(text), 'lightcyan'))
  # ====== training files ====== #
  print("#Train files:", ctext('%-8d' % len(train_indices), 'cyan'),
        "#spk:", ctext(len(set(name2label[name]
                               for name in train_indices.keys())), 'cyan'),
        "#noise:", ctext(len([name for name in train_indices.keys()
                              if '/' in name]), 'cyan'))
  summary_indices(ids=train_indices)
  # ====== valid files ====== #
  print("#Valid files:", ctext('%-8d' % len(valid_indices), 'cyan'),
        "#spk:", ctext(len(set(name2label[name]
                               for name in valid_indices.keys())), 'cyan'),
        "#noise:", ctext(len([name for name in valid_indices.keys()
                              if '/' in name]), 'cyan'))
  summary_indices(ids=valid_indices)
  # ******************** create the recipe ******************** #
  assert all(name in name2label
             for name in train_indices.keys())
  assert all(name in name2label
            for name in valid_indices.keys())
  recipes = prepare_dnn_feeder_recipe(name2label=name2label,
                                      n_speakers=len(all_speakers),
                                      utt_length=utt_length, seq_mode=seq_mode)
  # ====== downsample training set for analyzing if required ====== #
  if train_proportion is not None:
    assert 0 < train_proportion < 1
    n_training = len(train_indices)
    train_indices = list(train_indices.items())
    rand.shuffle(train_indices); rand.shuffle(train_indices)
    train_indices = dict(train_indices[:int(n_training * train_proportion)])
  # ====== create feeder ====== #
  train_feeder = F.Feeder(
      data_desc=F.IndexedData(data=X,
                              indices=train_indices),
      batch_mode='batch', ncpu=NCPU, buffer_size=256)

  valid_feeder = F.Feeder(
      data_desc=F.IndexedData(data=X,
                              indices=valid_indices),
      batch_mode='batch', ncpu=max(2, NCPU // 4), buffer_size=64)

  train_feeder.set_recipes(recipes)
  valid_feeder.set_recipes(recipes)
  print(train_feeder)
  print(valid_feeder)
  # ====== debugging ====== #
  if IS_DEBUGGING:
    import matplotlib
    matplotlib.use('Agg')
    prog = Progbar(target=len(valid_feeder), print_summary=True,
                   name="Iterating validation set")
    samples = []
    n_visual = 250
    for name, idx, X, y in valid_feeder.set_batch(batch_size=100000,
                                                  batch_mode='file',
                                                  seed=None, shuffle_level=0):
      assert idx == 0, "Utterances longer than %.2f(sec)" % (100000 * Config.STEP_LENGTH)
      prog['X'] = X.shape
      prog['y'] = y.shape
      prog.add(X.shape[0])
      # random sampling
      if rand.rand(1) < 0.5 and len(samples) < n_visual:
        for i in rand.randint(0, X.shape[0], size=4, dtype='int32'):
          samples.append((name, X[i], np.argmax(y[i], axis=-1)))
    # plot the spectrogram
    n_visual = len(samples)
    V.plot_figure(nrow=n_visual, ncol=8)
    for i, (name, X, y) in enumerate(samples):
      is_noise = '/' in name
      assert name2label[name] == y, "Speaker label mismatch for file: %s" % name
      name = name.split('/')[0]
      dsname = ds['dsname'][name]
      spkid = ds['spkid'][name]
      y = np.argmax(y, axis=-1)
      ax = V.plot_spectrogram(X.T,
                              ax=(n_visual, 1, i + 1),
                              title='#%d' % (i + 1))
      ax.set_title('[%s][%s]%s  %s' %
                   ('noise' if is_noise else 'clean', dsname, name, spkid),
                   fontsize=6)
    # don't need to be high resolutions
    V.plot_save('/tmp/tmp.pdf', dpi=12)
    exit()
  # ====== return ====== #
  if bool(return_dataset):
    return train_feeder, valid_feeder, all_speakers, ds
  return train_feeder, valid_feeder, all_speakers

# ===========================================================================
# Evaluation and validation helper
# ===========================================================================
def validate_features_dataset(output_dataset_path, ds_validation_path):
  ds = F.Dataset(output_dataset_path, read_only=True)
  print(ds)

  features = {}
  for key, val in ds.items():
    if 'indices_' in key:
      name = key.split('_')[-1]
      features[name] = (val, ds[name])

  all_indices = [val[0] for val in features.values()]
  # ====== sampling 250 files ====== #
  all_files = sampling_iter(it=all_indices[0].keys(), k=250,
                            seed=Config.SUPER_SEED)
  all_files = [f for f in all_files
               if all(f in ids for ids in all_indices)]
  print("#Samples:", ctext(len(all_files), 'cyan'))

  # ====== ignore the 20-figures warning ====== #
  with catch_warnings_ignore(RuntimeWarning):
    for file_name in all_files:
      X = {}
      for feat_name, (ids, data) in features.items():
        start, end = ids[file_name]
        X[feat_name] = data[start:end][:].astype('float32')
      V.plot_multiple_features(features=X, fig_width=20,
            title='[%s]%s' % (ds['dsname'][file_name], file_name))

  V.plot_save(ds_validation_path, dpi=12)



from os import path

from setuptools import find_packages, setup

_ODIN_VERSION_ = '1.2.0'

_TENSORFLOW_VERSION = '2.0.0'
_TENSORFLOW_PROBABILITY_VERSION = '0.8.0'
_PYTORCH_VERSION = '1.2.0'


# ===========================================================================
# Packages check
# ===========================================================================
def get_tensorflow_version():
  import subprocess
  try:
    task = subprocess.Popen(["nvcc", "--version"],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    out = task.stdout.read()
    if "release 9.0" in str(out, 'utf-8'):
      return "tensorflow-gpu==%s" % _TENSORFLOW_VERSION
  except FileNotFoundError:
    pass
  return "tensorflow==%s" % _TENSORFLOW_VERSION


# ===========================================================================
# Description
# ===========================================================================
here = path.abspath(path.dirname(__file__))

long_description = \
'''
An end-to-end framework support multi-modal data processing
and fast prototyping of machine learning algorithm in form
of organized networks.
'''

# ===========================================================================
# Setup
# ===========================================================================
setup(
    name='odin-ai',
    version=_ODIN_VERSION_,
    description="Deep learning for research and production",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/imito/odin-ai',
    author='Trung Ngo Trong',
    author_email='trung@imito.ai',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
    ],
    keywords=
    'tensorflow pytorch machine learning neural networks deep learning bayesian',
    packages=find_packages(exclude=['examples', 'examples/*', 'docs', 'tests']),
    # scripts=['bin/speech-augmentation', 'bin/speech-test'],
    setup_requires=['pip>=19.0'],
    install_requires=[
        'numpy>=1.0.0',
        get_tensorflow_version(),
        'tensorflow-probability==%s' % _TENSORFLOW_PROBABILITY_VERSION,
        'tensorflow-datasets',
        'torch==%s' % _PYTORCH_VERSION, 'six>=1.9.0', 'scikit-learn>=0.18.0',
        'matplotlib>=3.0.0', 'decorator', 'tqdm', 'dill', 'pyyaml', 'pycrypto'
    ],
    extras_require={
        'visualize': ['pydot>=1.2.4', 'colorama', 'seaborn'],
        'tests': ['pytest', 'pandas', 'requests'],
        'audio': ['soundfile', 'resampy'],
        'docs': ['sphinx', 'sphinx_rtd_theme']
    },
    zip_safe=False)
